# coding: latin-1

import uuid
from datetime import datetime
from json import loads
from sys import exc_info
from unittest import TestCase

from requests import get

from ocmpy import *


class ParseReferenceDataTests(TestCase):
    def test_ParseValidReferenceData(self):
        # Arrange
        testdata = loads(r'''{
          "ChargerTypes": [
            {
              "Comments": "Under 2 kW, usually domestic socket types",
              "IsFastChargeCapable": false,
              "ID": 1,
              "Title": "Level 1 : Low (Under 2kW)"
            },
            {
              "Comments": "Over 2 kW, usually non-domestic socket type",
              "IsFastChargeCapable": false,
              "ID": 2,
              "Title": "Level 2 : Medium (Over 2kW)"
            },
            {
              "Comments": "40KW and Higher",
              "IsFastChargeCapable": true,
              "ID": 3,
              "Title": "Level 3:  High (Over 40kW)"
            }
          ],
          "ConnectionTypes": [
            {
              "FormalName": "Not Specified",
              "IsDiscontinued": null,
              "IsObsolete": null,
              "ID": 0,
              "Title": "Unknown"
            },
            {
              "FormalName": "SAE J1772-2009",
              "IsDiscontinued": null,
              "IsObsolete": null,
              "ID": 1,
              "Title": "J1772"
            },
            {
              "FormalName": null,
              "IsDiscontinued": null,
              "IsObsolete": null,
              "ID": 2,
              "Title": "CHAdeMO"
            },
            {
              "FormalName": "BS1363 / Type G",
              "IsDiscontinued": null,
              "IsObsolete": null,
              "ID": 3,
              "Title": "BS1363 3 Pin 13 Amp"
            },
            {
              "FormalName": null,
              "IsDiscontinued": null,
              "IsObsolete": null,
              "ID": 4,
              "Title": "Blue Commando (2P+E)"
            },
            {
              "FormalName": "Large Paddle Inductive",
              "IsDiscontinued": true,
              "IsObsolete": true,
              "ID": 5,
              "Title": "LP Inductive"
            },
            {
              "FormalName": "Small Paddle Inductive",
              "IsDiscontinued": true,
              "IsObsolete": true,
              "ID": 6,
              "Title": "SP Inductive"
            },
            {
              "FormalName": "Avcon SAE J1772-2001",
              "IsDiscontinued": true,
              "IsObsolete": false,
              "ID": 7,
              "Title": "Avcon Connector"
            },
            {
              "FormalName": "Tesla Connector",
              "IsDiscontinued": true,
              "IsObsolete": false,
              "ID": 8,
              "Title": "Tesla (Roadster)"
            },
            {
              "FormalName": null,
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 9,
              "Title": "NEMA 5-20R"
            },
            {
              "FormalName": null,
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 10,
              "Title": "NEMA 14-30"
            },
            {
              "FormalName": null,
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 11,
              "Title": "NEMA 14-50"
            },
            {
              "FormalName": "Europlug 2-Pin (CEE 7/16)",
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 13,
              "Title": "Europlug 2-Pin (CEE 7/16)"
            },
            {
              "FormalName": null,
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 14,
              "Title": "NEMA 6-20"
            },
            {
              "FormalName": null,
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 15,
              "Title": "NEMA 6-15"
            },
            {
              "FormalName": null,
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 16,
              "Title": "CEE 3 Pin"
            },
            {
              "FormalName": null,
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 17,
              "Title": "CEE 5 Pin"
            },
            {
              "FormalName": null,
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 18,
              "Title": "CEE+ 7 Pin"
            },
            {
              "FormalName": null,
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 21,
              "Title": "XLR Plug (4 pin)"
            },
            {
              "FormalName": "NEMA 5-15R",
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 22,
              "Title": "NEMA 5-15R"
            },
            {
              "FormalName": null,
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 23,
              "Title": "CEE 7/5"
            },
            {
              "FormalName": null,
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 24,
              "Title": "Wireless Charging"
            },
            {
              "FormalName": "IEC 62196-2 Type 2",
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 25,
              "Title": "Mennekes (Type 2)"
            },
            {
              "FormalName": "IEC 62196-2 Type 3",
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 26,
              "Title": "SCAME Type 3C (Schneider-Legrand)"
            },
            {
              "FormalName": "Tesla Supercharger",
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 27,
              "Title": "Tesla Supercharger"
            },
            {
              "FormalName": "CEE 7/4",
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 28,
              "Title": "CEE 7/4 - Schuko - Type F"
            },
            {
              "FormalName": "Type I/AS 3112/CPCS-CCC",
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 29,
              "Title": "Type I (AS 3112)"
            },
            {
              "FormalName": null,
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 30,
              "Title": "Tesla (Model S onwards)"
            },
            {
              "FormalName": "Tesla Battery Swap Station",
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 31,
              "Title": "Tesla Battery Swap"
            },
            {
              "FormalName": null,
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 32,
              "Title": "SAE Combo (DC Fast Charge J1772)"
            },
            {
              "FormalName": null,
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 33,
              "Title": "CCS (Type 2 Version of Combined Coupler)"
            },
            {
              "FormalName": "IEC 60309 3-pin",
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 34,
              "Title": "IEC 60309 3-pin"
            },
            {
              "FormalName": "IEC 60309 5-pin",
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 35,
              "Title": "IEC 60309 5-pin"
            },
            {
              "FormalName": null,
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 36,
              "Title": "SCAME Type 3A (Low Power)"
            },
            {
              "FormalName": null,
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 1036,
              "Title": "Mennekes (Type 2, Tethered Connector) "
            },
            {
              "FormalName": "T13/ IEC Type J",
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 1037,
              "Title": "T13 - SEC1011 ( Swiss domestic 3-pin ) - Type J"
            },
            {
              "FormalName": "GB-T AC - GB/T 20234.2 (Socket)",
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 1038,
              "Title": "GB-T AC - GB/T 20234.2 (Socket)"
            },
            {
              "FormalName": "GB-T AC - GB/T 20234.2 (Tethered Cable)",
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 1039,
              "Title": "GB-T AC - GB/T 20234.2 (Tethered Cable)"
            },
            {
              "FormalName": "GB-T DC - GB/T 20234.3",
              "IsDiscontinued": false,
              "IsObsolete": false,
              "ID": 1040,
              "Title": "GB-T DC - GB/T 20234.3"
            }
          ],
          "CurrentTypes": [
            {
              "Description": "Alternating Current - Single Phase",
              "ID": 10,
              "Title": "AC (Single-Phase)"
            },
            {
              "Description": "Alternating Current - Three Phase",
              "ID": 20,
              "Title": "AC (Three-Phase)"
            },
            {
              "Description": "Direct Current",
              "ID": 30,
              "Title": "DC"
            }
          ],
          "Countries": [
            {
              "ISOCode": "GB",
              "ContinentCode": "EU",
              "ID": 1,
              "Title": "United Kingdom"
            },
            {
              "ISOCode": "US",
              "ContinentCode": "NA",
              "ID": 2,
              "Title": "United States"
            },
            {
              "ISOCode": "IE",
              "ContinentCode": "EU",
              "ID": 3,
              "Title": "Ireland"
            },
            {
              "ISOCode": "HK",
              "ContinentCode": "AS",
              "ID": 4,
              "Title": "Hong Kong"
            },
            {
              "ISOCode": "AF",
              "ContinentCode": "AS",
              "ID": 5,
              "Title": "Afghanistan"
            },
            {
              "ISOCode": "AX",
              "ContinentCode": "EU",
              "ID": 6,
              "Title": "Aland Islands"
            },
            {
              "ISOCode": "AL",
              "ContinentCode": "EU",
              "ID": 7,
              "Title": "Albania"
            },
            {
              "ISOCode": "DZ",
              "ContinentCode": "AF",
              "ID": 8,
              "Title": "Algeria"
            },
            {
              "ISOCode": "AS",
              "ContinentCode": "OC",
              "ID": 9,
              "Title": "American Samoa"
            },
            {
              "ISOCode": "AD",
              "ContinentCode": "EU",
              "ID": 10,
              "Title": "Andorra"
            },
            {
              "ISOCode": "AO",
              "ContinentCode": "AF",
              "ID": 11,
              "Title": "Angola"
            },
            {
              "ISOCode": "AI",
              "ContinentCode": "NA",
              "ID": 12,
              "Title": "Anguilla"
            },
            {
              "ISOCode": "AQ",
              "ContinentCode": "AN",
              "ID": 13,
              "Title": "Antarctica"
            },
            {
              "ISOCode": "AG",
              "ContinentCode": "NA",
              "ID": 14,
              "Title": "Antigua And Barbuda"
            },
            {
              "ISOCode": "AR",
              "ContinentCode": "SA",
              "ID": 15,
              "Title": "Argentina"
            },
            {
              "ISOCode": "AM",
              "ContinentCode": "AS",
              "ID": 16,
              "Title": "Armenia"
            },
            {
              "ISOCode": "AW",
              "ContinentCode": "NA",
              "ID": 17,
              "Title": "Aruba"
            },
            {
              "ISOCode": "AU",
              "ContinentCode": "OC",
              "ID": 18,
              "Title": "Australia"
            },
            {
              "ISOCode": "AT",
              "ContinentCode": "EU",
              "ID": 19,
              "Title": "Austria"
            },
            {
              "ISOCode": "AZ",
              "ContinentCode": "AS",
              "ID": 20,
              "Title": "Azerbaijan"
            },
            {
              "ISOCode": "BS",
              "ContinentCode": "NA",
              "ID": 21,
              "Title": "Bahamas"
            },
            {
              "ISOCode": "BH",
              "ContinentCode": "AS",
              "ID": 22,
              "Title": "Bahrain"
            },
            {
              "ISOCode": "BD",
              "ContinentCode": "AS",
              "ID": 23,
              "Title": "Bangladesh"
            },
            {
              "ISOCode": "BB",
              "ContinentCode": "NA",
              "ID": 24,
              "Title": "Barbados"
            },
            {
              "ISOCode": "BY",
              "ContinentCode": "EU",
              "ID": 25,
              "Title": "Belarus"
            },
            {
              "ISOCode": "BE",
              "ContinentCode": "EU",
              "ID": 26,
              "Title": "Belgium"
            },
            {
              "ISOCode": "BZ",
              "ContinentCode": "NA",
              "ID": 27,
              "Title": "Belize"
            },
            {
              "ISOCode": "BJ",
              "ContinentCode": "AF",
              "ID": 28,
              "Title": "Benin"
            },
            {
              "ISOCode": "BM",
              "ContinentCode": "NA",
              "ID": 29,
              "Title": "Bermuda"
            },
            {
              "ISOCode": "BT",
              "ContinentCode": "AS",
              "ID": 30,
              "Title": "Bhutan"
            },
            {
              "ISOCode": "BO",
              "ContinentCode": "SA",
              "ID": 31,
              "Title": "Bolivia, Plurinational State Of"
            },
            {
              "ISOCode": "BQ",
              "ContinentCode": null,
              "ID": 32,
              "Title": "Bonaire, Saint Eustatius And Saba"
            },
            {
              "ISOCode": "BA",
              "ContinentCode": "EU",
              "ID": 33,
              "Title": "Bosnia And Herzegovina"
            },
            {
              "ISOCode": "BW",
              "ContinentCode": "AF",
              "ID": 34,
              "Title": "Botswana"
            },
            {
              "ISOCode": "BV",
              "ContinentCode": "AN",
              "ID": 35,
              "Title": "Bouvet Island"
            },
            {
              "ISOCode": "BR",
              "ContinentCode": "SA",
              "ID": 36,
              "Title": "Brazil"
            },
            {
              "ISOCode": "IO",
              "ContinentCode": "AS",
              "ID": 37,
              "Title": "British Indian Ocean Territory"
            },
            {
              "ISOCode": "BN",
              "ContinentCode": "AS",
              "ID": 38,
              "Title": "Brunei Darussalam"
            },
            {
              "ISOCode": "BG",
              "ContinentCode": "EU",
              "ID": 39,
              "Title": "Bulgaria"
            },
            {
              "ISOCode": "BF",
              "ContinentCode": "AF",
              "ID": 40,
              "Title": "Burkina Faso"
            },
            {
              "ISOCode": "BI",
              "ContinentCode": "AF",
              "ID": 41,
              "Title": "Burundi"
            },
            {
              "ISOCode": "KH",
              "ContinentCode": "AS",
              "ID": 42,
              "Title": "Cambodia"
            },
            {
              "ISOCode": "CM",
              "ContinentCode": "AF",
              "ID": 43,
              "Title": "Cameroon"
            },
            {
              "ISOCode": "CA",
              "ContinentCode": "NA",
              "ID": 44,
              "Title": "Canada"
            },
            {
              "ISOCode": "CV",
              "ContinentCode": "AF",
              "ID": 45,
              "Title": "Cape Verde"
            },
            {
              "ISOCode": "KY",
              "ContinentCode": "NA",
              "ID": 46,
              "Title": "Cayman Islands"
            },
            {
              "ISOCode": "CF",
              "ContinentCode": "AF",
              "ID": 47,
              "Title": "Central African Republic"
            },
            {
              "ISOCode": "TD",
              "ContinentCode": "AF",
              "ID": 48,
              "Title": "Chad"
            },
            {
              "ISOCode": "CL",
              "ContinentCode": "SA",
              "ID": 49,
              "Title": "Chile"
            },
            {
              "ISOCode": "CN",
              "ContinentCode": "AS",
              "ID": 50,
              "Title": "China"
            },
            {
              "ISOCode": "CX",
              "ContinentCode": "AS",
              "ID": 51,
              "Title": "Christmas Island"
            },
            {
              "ISOCode": "CC",
              "ContinentCode": "AS",
              "ID": 52,
              "Title": "Cocos (Keeling) Islands"
            },
            {
              "ISOCode": "CO",
              "ContinentCode": "SA",
              "ID": 53,
              "Title": "Colombia"
            },
            {
              "ISOCode": "KM",
              "ContinentCode": "AF",
              "ID": 54,
              "Title": "Comoros"
            },
            {
              "ISOCode": "CG",
              "ContinentCode": "AF",
              "ID": 55,
              "Title": "Congo"
            },
            {
              "ISOCode": "CD",
              "ContinentCode": "AF",
              "ID": 56,
              "Title": "Congo, Democratic Republic Of The"
            },
            {
              "ISOCode": "CK",
              "ContinentCode": "OC",
              "ID": 57,
              "Title": "Cook Islands"
            },
            {
              "ISOCode": "CR",
              "ContinentCode": "NA",
              "ID": 58,
              "Title": "Costa Rica"
            },
            {
              "ISOCode": "CI",
              "ContinentCode": "AF",
              "ID": 59,
              "Title": "Cote D'Ivoire"
            },
            {
              "ISOCode": "HR",
              "ContinentCode": "EU",
              "ID": 60,
              "Title": "Croatia"
            },
            {
              "ISOCode": "CU",
              "ContinentCode": "NA",
              "ID": 61,
              "Title": "Cuba"
            },
            {
              "ISOCode": "CW",
              "ContinentCode": null,
              "ID": 62,
              "Title": "Curacao"
            },
            {
              "ISOCode": "CY",
              "ContinentCode": "AS",
              "ID": 63,
              "Title": "Cyprus"
            },
            {
              "ISOCode": "CZ",
              "ContinentCode": "EU",
              "ID": 64,
              "Title": "Czech Republic"
            },
            {
              "ISOCode": "DK",
              "ContinentCode": "EU",
              "ID": 65,
              "Title": "Denmark"
            },
            {
              "ISOCode": "DJ",
              "ContinentCode": "AF",
              "ID": 66,
              "Title": "Djibouti"
            },
            {
              "ISOCode": "DM",
              "ContinentCode": "NA",
              "ID": 67,
              "Title": "Dominica"
            },
            {
              "ISOCode": "DO",
              "ContinentCode": "NA",
              "ID": 68,
              "Title": "Dominican Republic"
            },
            {
              "ISOCode": "EC",
              "ContinentCode": "SA",
              "ID": 69,
              "Title": "Ecuador"
            },
            {
              "ISOCode": "EG",
              "ContinentCode": "AF",
              "ID": 70,
              "Title": "Egypt"
            },
            {
              "ISOCode": "SV",
              "ContinentCode": "NA",
              "ID": 71,
              "Title": "El Salvador"
            },
            {
              "ISOCode": "GQ",
              "ContinentCode": "AF",
              "ID": 72,
              "Title": "Equatorial Guinea"
            },
            {
              "ISOCode": "ER",
              "ContinentCode": "AF",
              "ID": 73,
              "Title": "Eritrea"
            },
            {
              "ISOCode": "EE",
              "ContinentCode": "EU",
              "ID": 74,
              "Title": "Estonia"
            },
            {
              "ISOCode": "ET",
              "ContinentCode": "AF",
              "ID": 75,
              "Title": "Ethiopia"
            },
            {
              "ISOCode": "FK",
              "ContinentCode": "SA",
              "ID": 76,
              "Title": "Falkland Islands (Malvinas)"
            },
            {
              "ISOCode": "FO",
              "ContinentCode": "EU",
              "ID": 77,
              "Title": "Faroe Islands"
            },
            {
              "ISOCode": "FJ",
              "ContinentCode": "OC",
              "ID": 78,
              "Title": "Fiji"
            },
            {
              "ISOCode": "FI",
              "ContinentCode": "EU",
              "ID": 79,
              "Title": "Finland"
            },
            {
              "ISOCode": "FR",
              "ContinentCode": "EU",
              "ID": 80,
              "Title": "France"
            },
            {
              "ISOCode": "GF",
              "ContinentCode": "SA",
              "ID": 81,
              "Title": "French Guiana"
            },
            {
              "ISOCode": "PF",
              "ContinentCode": "OC",
              "ID": 82,
              "Title": "French Polynesia"
            },
            {
              "ISOCode": "TF",
              "ContinentCode": "AN",
              "ID": 83,
              "Title": "French Southern Territories"
            },
            {
              "ISOCode": "GA",
              "ContinentCode": "AF",
              "ID": 84,
              "Title": "Gabon"
            },
            {
              "ISOCode": "GM",
              "ContinentCode": "AF",
              "ID": 85,
              "Title": "Gambia"
            },
            {
              "ISOCode": "GE",
              "ContinentCode": "AS",
              "ID": 86,
              "Title": "Georgia"
            },
            {
              "ISOCode": "DE",
              "ContinentCode": "EU",
              "ID": 87,
              "Title": "Germany"
            },
            {
              "ISOCode": "GH",
              "ContinentCode": "AF",
              "ID": 88,
              "Title": "Ghana"
            },
            {
              "ISOCode": "GI",
              "ContinentCode": "EU",
              "ID": 89,
              "Title": "Gibraltar"
            },
            {
              "ISOCode": "GR",
              "ContinentCode": "EU",
              "ID": 90,
              "Title": "Greece"
            },
            {
              "ISOCode": "GL",
              "ContinentCode": "NA",
              "ID": 91,
              "Title": "Greenland"
            },
            {
              "ISOCode": "GD",
              "ContinentCode": "NA",
              "ID": 92,
              "Title": "Grenada"
            },
            {
              "ISOCode": "GP",
              "ContinentCode": "NA",
              "ID": 93,
              "Title": "Guadeloupe"
            },
            {
              "ISOCode": "GU",
              "ContinentCode": "OC",
              "ID": 94,
              "Title": "Guam"
            },
            {
              "ISOCode": "GT",
              "ContinentCode": "NA",
              "ID": 95,
              "Title": "Guatemala"
            },
            {
              "ISOCode": "GG",
              "ContinentCode": "EU",
              "ID": 96,
              "Title": "Guernsey"
            },
            {
              "ISOCode": "GN",
              "ContinentCode": "AF",
              "ID": 97,
              "Title": "Guinea"
            },
            {
              "ISOCode": "GW",
              "ContinentCode": "AF",
              "ID": 98,
              "Title": "Guinea-Bissau"
            },
            {
              "ISOCode": "GY",
              "ContinentCode": "SA",
              "ID": 99,
              "Title": "Guyana"
            },
            {
              "ISOCode": "HT",
              "ContinentCode": "NA",
              "ID": 100,
              "Title": "Haiti"
            },
            {
              "ISOCode": "HM",
              "ContinentCode": "AN",
              "ID": 101,
              "Title": "Heard Island And Mcdonald Islands"
            },
            {
              "ISOCode": "VA",
              "ContinentCode": "EU",
              "ID": 102,
              "Title": "Holy See (Vatican City State)"
            },
            {
              "ISOCode": "HN",
              "ContinentCode": "NA",
              "ID": 103,
              "Title": "Honduras"
            },
            {
              "ISOCode": "HU",
              "ContinentCode": "EU",
              "ID": 104,
              "Title": "Hungary"
            },
            {
              "ISOCode": "IS",
              "ContinentCode": "EU",
              "ID": 105,
              "Title": "Iceland"
            },
            {
              "ISOCode": "IN",
              "ContinentCode": "AS",
              "ID": 106,
              "Title": "India"
            },
            {
              "ISOCode": "ID",
              "ContinentCode": "AS",
              "ID": 107,
              "Title": "Indonesia"
            },
            {
              "ISOCode": "IR",
              "ContinentCode": "AS",
              "ID": 108,
              "Title": "Iran, Islamic Republic Of"
            },
            {
              "ISOCode": "IQ",
              "ContinentCode": "AS",
              "ID": 109,
              "Title": "Iraq"
            },
            {
              "ISOCode": "IM",
              "ContinentCode": "EU",
              "ID": 110,
              "Title": "Isle Of Man"
            },
            {
              "ISOCode": "IL",
              "ContinentCode": "AS",
              "ID": 111,
              "Title": "Israel"
            },
            {
              "ISOCode": "IT",
              "ContinentCode": "EU",
              "ID": 112,
              "Title": "Italy"
            },
            {
              "ISOCode": "JM",
              "ContinentCode": "NA",
              "ID": 113,
              "Title": "Jamaica"
            },
            {
              "ISOCode": "JP",
              "ContinentCode": "AS",
              "ID": 114,
              "Title": "Japan"
            },
            {
              "ISOCode": "JE",
              "ContinentCode": "EU",
              "ID": 115,
              "Title": "Jersey"
            },
            {
              "ISOCode": "JO",
              "ContinentCode": "AS",
              "ID": 116,
              "Title": "Jordan"
            },
            {
              "ISOCode": "KZ",
              "ContinentCode": "AS",
              "ID": 117,
              "Title": "Kazakhstan"
            },
            {
              "ISOCode": "KE",
              "ContinentCode": "AF",
              "ID": 118,
              "Title": "Kenya"
            },
            {
              "ISOCode": "KI",
              "ContinentCode": "OC",
              "ID": 119,
              "Title": "Kiribati"
            },
            {
              "ISOCode": "KP",
              "ContinentCode": "AS",
              "ID": 120,
              "Title": "Korea, Democratic People's Rep."
            },
            {
              "ISOCode": "KR",
              "ContinentCode": "AS",
              "ID": 121,
              "Title": "Korea, Republic Of"
            },
            {
              "ISOCode": "KW",
              "ContinentCode": "AS",
              "ID": 122,
              "Title": "Kuwait"
            },
            {
              "ISOCode": "KG",
              "ContinentCode": "AS",
              "ID": 123,
              "Title": "Kyrgyzstan"
            },
            {
              "ISOCode": "LA",
              "ContinentCode": "AS",
              "ID": 124,
              "Title": "Lao People's Democratic Republic"
            },
            {
              "ISOCode": "LV",
              "ContinentCode": "EU",
              "ID": 125,
              "Title": "Latvia"
            },
            {
              "ISOCode": "LB",
              "ContinentCode": "AS",
              "ID": 126,
              "Title": "Lebanon"
            },
            {
              "ISOCode": "LS",
              "ContinentCode": "AF",
              "ID": 127,
              "Title": "Lesotho"
            },
            {
              "ISOCode": "LR",
              "ContinentCode": "AF",
              "ID": 128,
              "Title": "Liberia"
            },
            {
              "ISOCode": "LY",
              "ContinentCode": "AF",
              "ID": 129,
              "Title": "Libyan Arab Jamahiriya"
            },
            {
              "ISOCode": "LI",
              "ContinentCode": "EU",
              "ID": 130,
              "Title": "Liechtenstein"
            },
            {
              "ISOCode": "LT",
              "ContinentCode": "EU",
              "ID": 131,
              "Title": "Lithuania"
            },
            {
              "ISOCode": "LU",
              "ContinentCode": "EU",
              "ID": 132,
              "Title": "Luxembourg"
            },
            {
              "ISOCode": "MO",
              "ContinentCode": "AS",
              "ID": 133,
              "Title": "Macao"
            },
            {
              "ISOCode": "MK",
              "ContinentCode": "EU",
              "ID": 134,
              "Title": "Macedonia"
            },
            {
              "ISOCode": "MG",
              "ContinentCode": "AF",
              "ID": 135,
              "Title": "Madagascar"
            },
            {
              "ISOCode": "MW",
              "ContinentCode": "AF",
              "ID": 136,
              "Title": "Malawi"
            },
            {
              "ISOCode": "MY",
              "ContinentCode": "AS",
              "ID": 137,
              "Title": "Malaysia"
            },
            {
              "ISOCode": "MV",
              "ContinentCode": "AS",
              "ID": 138,
              "Title": "Maldives"
            },
            {
              "ISOCode": "ML",
              "ContinentCode": "AF",
              "ID": 139,
              "Title": "Mali"
            },
            {
              "ISOCode": "MT",
              "ContinentCode": "EU",
              "ID": 140,
              "Title": "Malta"
            },
            {
              "ISOCode": "MH",
              "ContinentCode": "OC",
              "ID": 141,
              "Title": "Marshall Islands"
            },
            {
              "ISOCode": "MQ",
              "ContinentCode": "NA",
              "ID": 142,
              "Title": "Martinique"
            },
            {
              "ISOCode": "MR",
              "ContinentCode": "AF",
              "ID": 143,
              "Title": "Mauritania"
            },
            {
              "ISOCode": "MU",
              "ContinentCode": "AF",
              "ID": 144,
              "Title": "Mauritius"
            },
            {
              "ISOCode": "YT",
              "ContinentCode": "AF",
              "ID": 145,
              "Title": "Mayotte"
            },
            {
              "ISOCode": "MX",
              "ContinentCode": "NA",
              "ID": 146,
              "Title": "Mexico"
            },
            {
              "ISOCode": "FM",
              "ContinentCode": "OC",
              "ID": 147,
              "Title": "Micronesia, Federated States Of"
            },
            {
              "ISOCode": "MD",
              "ContinentCode": "EU",
              "ID": 148,
              "Title": "Moldova, Republic Of"
            },
            {
              "ISOCode": "MC",
              "ContinentCode": "EU",
              "ID": 149,
              "Title": "Monaco"
            },
            {
              "ISOCode": "MN",
              "ContinentCode": "AS",
              "ID": 150,
              "Title": "Mongolia"
            },
            {
              "ISOCode": "ME",
              "ContinentCode": "EU",
              "ID": 151,
              "Title": "Montenegro"
            },
            {
              "ISOCode": "MS",
              "ContinentCode": "NA",
              "ID": 152,
              "Title": "Montserrat"
            },
            {
              "ISOCode": "MA",
              "ContinentCode": "AF",
              "ID": 153,
              "Title": "Morocco"
            },
            {
              "ISOCode": "MZ",
              "ContinentCode": "AF",
              "ID": 154,
              "Title": "Mozambique"
            },
            {
              "ISOCode": "MM",
              "ContinentCode": "AS",
              "ID": 155,
              "Title": "Myanmar"
            },
            {
              "ISOCode": "NA",
              "ContinentCode": "AF",
              "ID": 156,
              "Title": "Namibia"
            },
            {
              "ISOCode": "NR",
              "ContinentCode": "OC",
              "ID": 157,
              "Title": "Nauru"
            },
            {
              "ISOCode": "NP",
              "ContinentCode": "AS",
              "ID": 158,
              "Title": "Nepal"
            },
            {
              "ISOCode": "NL",
              "ContinentCode": "EU",
              "ID": 159,
              "Title": "Netherlands"
            },
            {
              "ISOCode": "NC",
              "ContinentCode": "OC",
              "ID": 160,
              "Title": "New Caledonia"
            },
            {
              "ISOCode": "NZ",
              "ContinentCode": "OC",
              "ID": 161,
              "Title": "New Zealand"
            },
            {
              "ISOCode": "NI",
              "ContinentCode": "NA",
              "ID": 162,
              "Title": "Nicaragua"
            },
            {
              "ISOCode": "NE",
              "ContinentCode": "AF",
              "ID": 163,
              "Title": "Niger"
            },
            {
              "ISOCode": "NG",
              "ContinentCode": "AF",
              "ID": 164,
              "Title": "Nigeria"
            },
            {
              "ISOCode": "NU",
              "ContinentCode": "OC",
              "ID": 165,
              "Title": "Niue"
            },
            {
              "ISOCode": "NF",
              "ContinentCode": "OC",
              "ID": 166,
              "Title": "Norfolk Island"
            },
            {
              "ISOCode": "MP",
              "ContinentCode": "OC",
              "ID": 167,
              "Title": "Northern Mariana Islands"
            },
            {
              "ISOCode": "NO",
              "ContinentCode": "EU",
              "ID": 168,
              "Title": "Norway"
            },
            {
              "ISOCode": "OM",
              "ContinentCode": "AS",
              "ID": 169,
              "Title": "Oman"
            },
            {
              "ISOCode": "PK",
              "ContinentCode": "AS",
              "ID": 170,
              "Title": "Pakistan"
            },
            {
              "ISOCode": "PW",
              "ContinentCode": "OC",
              "ID": 171,
              "Title": "Palau"
            },
            {
              "ISOCode": "PS",
              "ContinentCode": "AS",
              "ID": 172,
              "Title": "Palestinian Territory, Occupied"
            },
            {
              "ISOCode": "PA",
              "ContinentCode": "NA",
              "ID": 173,
              "Title": "Panama"
            },
            {
              "ISOCode": "PG",
              "ContinentCode": "OC",
              "ID": 174,
              "Title": "Papua New Guinea"
            },
            {
              "ISOCode": "PY",
              "ContinentCode": "SA",
              "ID": 175,
              "Title": "Paraguay"
            },
            {
              "ISOCode": "PE",
              "ContinentCode": "SA",
              "ID": 176,
              "Title": "Peru"
            },
            {
              "ISOCode": "PH",
              "ContinentCode": "AS",
              "ID": 177,
              "Title": "Philippines"
            },
            {
              "ISOCode": "PN",
              "ContinentCode": "OC",
              "ID": 178,
              "Title": "Pitcairn"
            },
            {
              "ISOCode": "PL",
              "ContinentCode": "EU",
              "ID": 179,
              "Title": "Poland"
            },
            {
              "ISOCode": "PT",
              "ContinentCode": "EU",
              "ID": 180,
              "Title": "Portugal"
            },
            {
              "ISOCode": "PR",
              "ContinentCode": "NA",
              "ID": 181,
              "Title": "Puerto Rico"
            },
            {
              "ISOCode": "QA",
              "ContinentCode": "AS",
              "ID": 182,
              "Title": "Qatar"
            },
            {
              "ISOCode": "RE",
              "ContinentCode": "AF",
              "ID": 183,
              "Title": "Reunion"
            },
            {
              "ISOCode": "RO",
              "ContinentCode": "EU",
              "ID": 184,
              "Title": "Romania"
            },
            {
              "ISOCode": "RU",
              "ContinentCode": "EU",
              "ID": 185,
              "Title": "Russian Federation"
            },
            {
              "ISOCode": "RW",
              "ContinentCode": "AF",
              "ID": 186,
              "Title": "Rwanda"
            },
            {
              "ISOCode": "BL",
              "ContinentCode": "NA",
              "ID": 187,
              "Title": "Saint Barthelemy"
            },
            {
              "ISOCode": "SH",
              "ContinentCode": "AF",
              "ID": 188,
              "Title": "Saint Helena"
            },
            {
              "ISOCode": "KN",
              "ContinentCode": "NA",
              "ID": 189,
              "Title": "Saint Kitts And Nevis"
            },
            {
              "ISOCode": "LC",
              "ContinentCode": "NA",
              "ID": 190,
              "Title": "Saint Lucia"
            },
            {
              "ISOCode": "MF",
              "ContinentCode": "NA",
              "ID": 191,
              "Title": "Saint Martin (French Part)"
            },
            {
              "ISOCode": "PM",
              "ContinentCode": "NA",
              "ID": 192,
              "Title": "Saint Pierre And Miquelon"
            },
            {
              "ISOCode": "VC",
              "ContinentCode": "NA",
              "ID": 193,
              "Title": "Saint Vincent And The Grenadines"
            },
            {
              "ISOCode": "WS",
              "ContinentCode": "OC",
              "ID": 194,
              "Title": "Samoa"
            },
            {
              "ISOCode": "SM",
              "ContinentCode": "EU",
              "ID": 195,
              "Title": "San Marino"
            },
            {
              "ISOCode": "ST",
              "ContinentCode": "AF",
              "ID": 196,
              "Title": "Sao Tome And Principe"
            },
            {
              "ISOCode": "SA",
              "ContinentCode": "AS",
              "ID": 197,
              "Title": "Saudi Arabia"
            },
            {
              "ISOCode": "SN",
              "ContinentCode": "AF",
              "ID": 198,
              "Title": "Senegal"
            },
            {
              "ISOCode": "RS",
              "ContinentCode": "EU",
              "ID": 199,
              "Title": "Serbia"
            },
            {
              "ISOCode": "SC",
              "ContinentCode": "AF",
              "ID": 200,
              "Title": "Seychelles"
            },
            {
              "ISOCode": "SL",
              "ContinentCode": "AF",
              "ID": 201,
              "Title": "Sierra Leone"
            },
            {
              "ISOCode": "SG",
              "ContinentCode": "AS",
              "ID": 202,
              "Title": "Singapore"
            },
            {
              "ISOCode": "SX",
              "ContinentCode": null,
              "ID": 203,
              "Title": "Sint Maarten (Dutch Part)"
            },
            {
              "ISOCode": "SK",
              "ContinentCode": "EU",
              "ID": 204,
              "Title": "Slovakia"
            },
            {
              "ISOCode": "SI",
              "ContinentCode": "EU",
              "ID": 205,
              "Title": "Slovenia"
            },
            {
              "ISOCode": "SB",
              "ContinentCode": "OC",
              "ID": 206,
              "Title": "Solomon Islands"
            },
            {
              "ISOCode": "SO",
              "ContinentCode": "AF",
              "ID": 207,
              "Title": "Somalia"
            },
            {
              "ISOCode": "ZA",
              "ContinentCode": "AF",
              "ID": 208,
              "Title": "South Africa"
            },
            {
              "ISOCode": "GS",
              "ContinentCode": "AN",
              "ID": 209,
              "Title": "South Georgia"
            },
            {
              "ISOCode": "ES",
              "ContinentCode": "EU",
              "ID": 210,
              "Title": "Spain"
            },
            {
              "ISOCode": "LK",
              "ContinentCode": "AS",
              "ID": 211,
              "Title": "Sri Lanka"
            },
            {
              "ISOCode": "SD",
              "ContinentCode": "AF",
              "ID": 212,
              "Title": "Sudan"
            },
            {
              "ISOCode": "SR",
              "ContinentCode": "SA",
              "ID": 213,
              "Title": "Suriname"
            },
            {
              "ISOCode": "SJ",
              "ContinentCode": "EU",
              "ID": 214,
              "Title": "Svalbard And Jan Mayen"
            },
            {
              "ISOCode": "SZ",
              "ContinentCode": "AF",
              "ID": 215,
              "Title": "Swaziland"
            },
            {
              "ISOCode": "SE",
              "ContinentCode": "EU",
              "ID": 216,
              "Title": "Sweden"
            },
            {
              "ISOCode": "CH",
              "ContinentCode": "EU",
              "ID": 217,
              "Title": "Switzerland"
            },
            {
              "ISOCode": "SY",
              "ContinentCode": "AS",
              "ID": 218,
              "Title": "Syrian Arab Republic"
            },
            {
              "ISOCode": "TW",
              "ContinentCode": "AS",
              "ID": 219,
              "Title": "Taiwan, Province Of China"
            },
            {
              "ISOCode": "TJ",
              "ContinentCode": "AS",
              "ID": 220,
              "Title": "Tajikistan"
            },
            {
              "ISOCode": "TZ",
              "ContinentCode": "AF",
              "ID": 221,
              "Title": "Tanzania, United Republic Of"
            },
            {
              "ISOCode": "TH",
              "ContinentCode": "AS",
              "ID": 222,
              "Title": "Thailand"
            },
            {
              "ISOCode": "TL",
              "ContinentCode": "AS",
              "ID": 223,
              "Title": "Timor-Leste"
            },
            {
              "ISOCode": "TG",
              "ContinentCode": "AF",
              "ID": 224,
              "Title": "Togo"
            },
            {
              "ISOCode": "TK",
              "ContinentCode": "OC",
              "ID": 225,
              "Title": "Tokelau"
            },
            {
              "ISOCode": "TO",
              "ContinentCode": "OC",
              "ID": 226,
              "Title": "Tonga"
            },
            {
              "ISOCode": "TT",
              "ContinentCode": "NA",
              "ID": 227,
              "Title": "Trinidad And Tobago"
            },
            {
              "ISOCode": "TN",
              "ContinentCode": "AF",
              "ID": 228,
              "Title": "Tunisia"
            },
            {
              "ISOCode": "TR",
              "ContinentCode": "EU",
              "ID": 229,
              "Title": "Turkey"
            },
            {
              "ISOCode": "TM",
              "ContinentCode": "AS",
              "ID": 230,
              "Title": "Turkmenistan"
            },
            {
              "ISOCode": "TC",
              "ContinentCode": "NA",
              "ID": 231,
              "Title": "Turks And Caicos Islands"
            },
            {
              "ISOCode": "TV",
              "ContinentCode": "OC",
              "ID": 232,
              "Title": "Tuvalu"
            },
            {
              "ISOCode": "UG",
              "ContinentCode": "AF",
              "ID": 233,
              "Title": "Uganda"
            },
            {
              "ISOCode": "UA",
              "ContinentCode": "EU",
              "ID": 234,
              "Title": "Ukraine"
            },
            {
              "ISOCode": "AE",
              "ContinentCode": "AS",
              "ID": 235,
              "Title": "United Arab Emirates"
            },
            {
              "ISOCode": "UM",
              "ContinentCode": "OC",
              "ID": 236,
              "Title": "United States Minor Outlying Islands"
            },
            {
              "ISOCode": "UY",
              "ContinentCode": "SA",
              "ID": 237,
              "Title": "Uruguay"
            },
            {
              "ISOCode": "UZ",
              "ContinentCode": "AS",
              "ID": 238,
              "Title": "Uzbekistan"
            },
            {
              "ISOCode": "VU",
              "ContinentCode": "OC",
              "ID": 239,
              "Title": "Vanuatu"
            },
            {
              "ISOCode": "VE",
              "ContinentCode": "SA",
              "ID": 240,
              "Title": "Venezuela, Bolivarian Republic Of"
            },
            {
              "ISOCode": "VN",
              "ContinentCode": "AS",
              "ID": 241,
              "Title": "Viet Nam"
            },
            {
              "ISOCode": "VG",
              "ContinentCode": "NA",
              "ID": 242,
              "Title": "Virgin Islands, British"
            },
            {
              "ISOCode": "VI",
              "ContinentCode": "NA",
              "ID": 243,
              "Title": "Virgin Islands, U.S."
            },
            {
              "ISOCode": "WF",
              "ContinentCode": "OC",
              "ID": 244,
              "Title": "Wallis And Futuna"
            },
            {
              "ISOCode": "EH",
              "ContinentCode": "AF",
              "ID": 245,
              "Title": "Western Sahara"
            },
            {
              "ISOCode": "YE",
              "ContinentCode": "AS",
              "ID": 246,
              "Title": "Yemen"
            },
            {
              "ISOCode": "ZM",
              "ContinentCode": "AF",
              "ID": 247,
              "Title": "Zambia"
            },
            {
              "ISOCode": "ZW",
              "ContinentCode": "AF",
              "ID": 248,
              "Title": "Zimbabwe"
            },
            {
              "ISOCode": "SS",
              "ContinentCode": "AF",
              "ID": 249,
              "Title": "South Sudan"
            }
          ],
          "DataProviders": [
            {
              "WebsiteURL": "http://openchargemap.org",
              "Comments": null,
              "DataProviderStatusType": {
                "IsProviderEnabled": true,
                "ID": 1,
                "Title": "Manual Data Entry"
              },
              "IsRestrictedEdit": false,
              "IsOpenDataLicensed": true,
              "IsApprovedImport": true,
              "License": "Licensed under Creative Commons Attribution-ShareAlike 4.0 International (CC BY-SA 4.0)",
              "DateLastImported": null,
              "ID": 1,
              "Title": "Open Charge Map Contributors"
            },
            {
              "WebsiteURL": "http://www.afdc.energy.gov/",
              "Comments": null,
              "DataProviderStatusType": {
                "IsProviderEnabled": true,
                "ID": 20,
                "Title": "Automated Import"
              },
              "IsRestrictedEdit": false,
              "IsOpenDataLicensed": true,
              "IsApprovedImport": true,
              "License": "This data is provided by the National Renewable Energy Laboratory (\"NREL\"), which is operated by the Alliance for Sustainable Energy, LLC (\"Alliance\"), for the U.S. Department of Energy (\"DOE\"), and may be used for any purpose whatsoever.",
              "DateLastImported": "2018-04-21T04:28:19.807Z",
              "ID": 2,
              "Title": "afdc.energy.gov"
            },
            {
              "WebsiteURL": "http://www.mobie.pt",
              "Comments": null,
              "DataProviderStatusType": {
                "IsProviderEnabled": true,
                "ID": 20,
                "Title": "Automated Import"
              },
              "IsRestrictedEdit": false,
              "IsOpenDataLicensed": true,
              "IsApprovedImport": true,
              "License": "Public Data redistributed by agreement",
              "DateLastImported": "2017-10-28T13:32:10.85Z",
              "ID": 7,
              "Title": "Mobie.pt"
            },
            {
              "WebsiteURL": "http://e-tankstellen-finder.com/",
              "Comments": null,
              "DataProviderStatusType": {
                "IsProviderEnabled": true,
                "ID": 10,
                "Title": "Manual (Bulk Import)"
              },
              "IsRestrictedEdit": false,
              "IsOpenDataLicensed": null,
              "IsApprovedImport": null,
              "License": null,
              "DateLastImported": null,
              "ID": 8,
              "Title": "ev-charging.com"
            },
            {
              "WebsiteURL": "http://www.uppladdning.nu",
              "Comments": "This data provider has explicitly asked that OCM do not import their data. (Ref: Nikolay  Shishkov)",
              "DataProviderStatusType": {
                "IsProviderEnabled": false,
                "ID": 1000,
                "Title": "Delisted - Data Use Permission Denied"
              },
              "IsRestrictedEdit": false,
              "IsOpenDataLicensed": null,
              "IsApprovedImport": null,
              "License": null,
              "DateLastImported": null,
              "ID": 13,
              "Title": "www.uppladdning.nu"
            },
            {
              "WebsiteURL": "http://www.e-laad.nl",
              "Comments": null,
              "DataProviderStatusType": {
                "IsProviderEnabled": true,
                "ID": 10,
                "Title": "Manual (Bulk Import)"
              },
              "IsRestrictedEdit": false,
              "IsOpenDataLicensed": null,
              "IsApprovedImport": null,
              "License": null,
              "DateLastImported": null,
              "ID": 14,
              "Title": "e-Laad"
            },
            {
              "WebsiteURL": "http://www.carstations.com",
              "Comments": null,
              "DataProviderStatusType": {
                "IsProviderEnabled": true,
                "ID": 20,
                "Title": "Automated Import"
              },
              "IsRestrictedEdit": false,
              "IsOpenDataLicensed": null,
              "IsApprovedImport": null,
              "License": null,
              "DateLastImported": "2017-11-04T14:07:02.687Z",
              "ID": 15,
              "Title": "CarStations.com"
            },
            {
              "WebsiteURL": "http://www.blinknetwork.com",
              "Comments": null,
              "DataProviderStatusType": {
                "IsProviderEnabled": true,
                "ID": 20,
                "Title": "Automated Import"
              },
              "IsRestrictedEdit": false,
              "IsOpenDataLicensed": null,
              "IsApprovedImport": null,
              "License": null,
              "DateLastImported": null,
              "ID": 17,
              "Title": "BlinkNetwork.com"
            },
            {
              "WebsiteURL": "http://chargepoints.dft.gov.uk",
              "Comments": null,
              "DataProviderStatusType": {
                "IsProviderEnabled": true,
                "ID": 20,
                "Title": "Automated Import"
              },
              "IsRestrictedEdit": false,
              "IsOpenDataLicensed": true,
              "IsApprovedImport": true,
              "License": "Contains public sector information licensed under the Open Government Licence v2.0.",
              "DateLastImported": "2018-05-18T05:51:37.243Z",
              "ID": 18,
              "Title": "UK National Charge Point Registry"
            },
            {
              "WebsiteURL": "http://www.nobil.no",
              "Comments": "NOBIL is an Norwegian open, national database for registering and reporting information about charging stations.",
              "DataProviderStatusType": {
                "IsProviderEnabled": true,
                "ID": 30,
                "Title": "Partially Automated (Ad-Hoc Import)"
              },
              "IsRestrictedEdit": false,
              "IsOpenDataLicensed": true,
              "IsApprovedImport": true,
              "License": "NOBIL by Transnova, Norway is licensed under Creative Commons 3.0 Unported License",
              "DateLastImported": "2018-04-21T06:56:32.22Z",
              "ID": 19,
              "Title": "Nobil.no"
            },
            {
              "WebsiteURL": "http://www.chargepoint.net/",
              "Comments": "Coulomb Chargepoint Network",
              "DataProviderStatusType": {
                "IsProviderEnabled": false,
                "ID": 1001,
                "Title": "Delisted - Not Used"
              },
              "IsRestrictedEdit": false,
              "IsOpenDataLicensed": null,
              "IsApprovedImport": null,
              "License": null,
              "DateLastImported": null,
              "ID": 20,
              "Title": "Chargepoint.net"
            },
            {
              "WebsiteURL": "https://www.rwe-mobility.com/",
              "Comments": null,
              "DataProviderStatusType": {
                "IsProviderEnabled": true,
                "ID": 20,
                "Title": "Automated Import"
              },
              "IsRestrictedEdit": false,
              "IsOpenDataLicensed": null,
              "IsApprovedImport": null,
              "License": null,
              "DateLastImported": null,
              "ID": 21,
              "Title": "RWE Mobility"
            },
            {
              "WebsiteURL": "http://www.chademo.com",
              "Comments": null,
              "DataProviderStatusType": {
                "IsProviderEnabled": true,
                "ID": 100,
                "Title": "Not Currently Used/Imported"
              },
              "IsRestrictedEdit": false,
              "IsOpenDataLicensed": false,
              "IsApprovedImport": false,
              "License": null,
              "DateLastImported": null,
              "ID": 22,
              "Title": "CHAdeMO.com"
            },
            {
              "WebsiteURL": "http://www.esb.ie/electric-cars/",
              "Comments": null,
              "DataProviderStatusType": {
                "IsProviderEnabled": true,
                "ID": 20,
                "Title": "Automated Import"
              },
              "IsRestrictedEdit": false,
              "IsOpenDataLicensed": false,
              "IsApprovedImport": false,
              "License": null,
              "DateLastImported": "2016-05-29T07:34:20.087Z",
              "ID": 23,
              "Title": "ESB eCars"
            },
            {
              "WebsiteURL": "http://www.addenergietechnologies.com/",
              "Comments": null,
              "DataProviderStatusType": {
                "IsProviderEnabled": true,
                "ID": 30,
                "Title": "Partially Automated (Ad-Hoc Import)"
              },
              "IsRestrictedEdit": false,
              "IsOpenDataLicensed": true,
              "IsApprovedImport": true,
              "License": "http://creativecommons.org/licenses/by/3.0",
              "DateLastImported": "2017-09-07T10:50:55.78Z",
              "ID": 24,
              "Title": "AddEnergie Technologies Inc."
            },
            {
              "WebsiteURL": "http://icaen.gencat.cat/",
              "Comments": null,
              "DataProviderStatusType": {
                "IsProviderEnabled": true,
                "ID": 30,
                "Title": "Partially Automated (Ad-Hoc Import)"
              },
              "IsRestrictedEdit": false,
              "IsOpenDataLicensed": true,
              "IsApprovedImport": true,
              "License": "General public data",
              "DateLastImported": "2016-03-03T14:18:07.287Z",
              "ID": 25,
              "Title": "Catalan Energy Institute (ICAEN)"
            },
            {
              "WebsiteURL": "http://www.oplaadpalen.nl/",
              "Comments": null,
              "DataProviderStatusType": {
                "IsProviderEnabled": true,
                "ID": 20,
                "Title": "Automated Import"
              },
              "IsRestrictedEdit": false,
              "IsOpenDataLicensed": true,
              "IsApprovedImport": true,
              "License": "Licensed under Attribution-NonCommercial-ShareAlike 3.0 : http://creativecommons.org/licenses/by-nc-sa/3.0/",
              "DateLastImported": "2016-06-04T03:04:53.05Z",
              "ID": 26,
              "Title": "Oplaadpalen.nl"
            },
            {
              "WebsiteURL": "https://www.placetoplug.com",
              "Comments": null,
              "DataProviderStatusType": {
                "IsProviderEnabled": true,
                "ID": 30,
                "Title": "Partially Automated (Ad-Hoc Import)"
              },
              "IsRestrictedEdit": false,
              "IsOpenDataLicensed": false,
              "IsApprovedImport": false,
              "License": null,
              "DateLastImported": null,
              "ID": 27,
              "Title": "Place To Plug"
            }
          ],
          "Operators": [
            {
              "WebsiteURL": null,
              "Comments": "For use when the operator of the equipment is a single business owner connected to the location and equipment is not part of a larger network",
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 45,
              "Title": "(Business Owner at Location)"
            },
            {
              "WebsiteURL": null,
              "Comments": "For use when the operator is a home owner or private individual making their own facilities available to the public",
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": true,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 44,
              "Title": "(Private Residence/Individual)"
            },
            {
              "WebsiteURL": null,
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 1,
              "Title": "(Unknown Operator)"
            },
            {
              "WebsiteURL": "http://www.avinc.com/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 26,
              "Title": "AeroVironment"
            },
            {
              "WebsiteURL": "http://www.alfen.com/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 180,
              "Title": "Alfen"
            },
            {
              "WebsiteURL": "http://www.allego.eu/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@allego.eu",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 103,
              "Title": "Allego BV"
            },
            {
              "WebsiteURL": "https://www.alperiaenergy.eu/irj/portal/aew/e_mobility",
              "Comments": "formerly known as aew",
              "PhonePrimaryContact": "800110055",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "service@alperiaenergy.eu",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 164,
              "Title": "Alperia"
            },
            {
              "WebsiteURL": "http://www.amb.cat/en/web/mobilitat/mobilitat-sostenible/electrolineres",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3305,
              "Title": "AMB (rea metropolitana de Barcelona)"
            },
            {
              "WebsiteURL": "http://nou.amersam.cat/",
              "Comments": null,
              "PhonePrimaryContact": " +34 977 300 006",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "amersam@amersam.com",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 212,
              "Title": "Amersam"
            },
            {
              "WebsiteURL": "http://www.auto-bleue.org/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 56,
              "Title": "Auto Bleue"
            },
            {
              "WebsiteURL": "https://www.autolib.eu/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 53,
              "Title": "Autolib (Paris)"
            },
            {
              "WebsiteURL": "http://autopildyk.lt/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "pagalba@autopildyk.lt",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3284,
              "Title": "autoPilDYK"
            },
            {
              "WebsiteURL": "https://www.avacon.de/",
              "Comments": null,
              "PhonePrimaryContact": "+49 5351 1230",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@avacon.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3303,
              "Title": "Avacon"
            },
            {
              "WebsiteURL": null,
              "Comments": "Obsolete inductive paddle charging",
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 43,
              "Title": "AVCON"
            },
            {
              "WebsiteURL": "http://azranetwork.co/azra/",
              "Comments": null,
              "PhonePrimaryContact": "450 477-8008",
              "PhoneSecondaryContact": "1-844-279-2972",
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@azranetwork.co",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 186,
              "Title": "Azra Network"
            },
            {
              "WebsiteURL": "https://www.badenova.de/web/Privatkunden/E-Mobilit%C3%A4t/%C3%96ffentliches-Laden/index-2.jsp",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "e-mobility@badenova.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3292,
              "Title": "Badenova (DE)"
            },
            {
              "WebsiteURL": "http://www.be-emobil.de/",
              "Comments": null,
              "PhonePrimaryContact": "+49 30 20847590",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "dialog@be-emobil.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 214,
              "Title": "be emobil"
            },
            {
              "WebsiteURL": "https://beenergised.com/",
              "Comments": "also known as has.to.be",
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3261,
              "Title": "Be Energised"
            },
            {
              "WebsiteURL": "http://www.becharged.eu/",
              "Comments": null,
              "PhonePrimaryContact": "+32 (0)9 395.05.93",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@becharged.eu",
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 94,
              "Title": "BeCharged"
            },
            {
              "WebsiteURL": "http://belib.paris/",
              "Comments": null,
              "PhonePrimaryContact": "09.69.322.500 (7j/7 de 7h  22h)",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "contact@sodetrel.fr",
              "FaultReportEmail": "sav@sodetrel.fr",
              "IsRestrictedEdit": false,
              "ID": 1235,
              "Title": "Belib?"
            },
            {
              "WebsiteURL": "http://www.belorusneft.by/beloil-map/?lang=en",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3268,
              "Title": "Belorusneft"
            },
            {
              "WebsiteURL": "http://www.blinknetwork.com/",
              "Comments": "Membership based charging network, part of Car Charging Group, inc.",
              "PhonePrimaryContact": "888.998.BLINK (2546)",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": "https://prod.blinknetwork.com/membership.html",
              "ContactEmail": "support@blinknetwork.com",
              "FaultReportEmail": "support@blinknetwork.com",
              "IsRestrictedEdit": false,
              "ID": 9,
              "Title": "Blink Network/ECOtality"
            },
            {
              "WebsiteURL": "http://www.bluecorner.be/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 36,
              "Title": "Blue Corner (Belgium)"
            },
            {
              "WebsiteURL": "http://www.bs-energy.de/engagement/umwelt/elektromobilitaet/elektrotankstellen/",
              "Comments": null,
              "PhonePrimaryContact": " 0531 3838000",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "energieberatung@bs-energy.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 129,
              "Title": "BS Energie"
            },
            {
              "WebsiteURL": "http://carcharger.ie",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 1241,
              "Title": "Carcharger.ie"
            },
            {
              "WebsiteURL": "https://www.celsia.com/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3304,
              "Title": "Celsia"
            },
            {
              "WebsiteURL": "http://www.elektromobilita.cz/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 115,
              "Title": "?EZ"
            },
            {
              "WebsiteURL": "https://chargedrive.com/",
              "Comments": null,
              "PhonePrimaryContact": " +358 40 709 5566 ",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "rami.syvari@fortum.com",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 198,
              "Title": "Charge & Drive"
            },
            {
              "WebsiteURL": "http://www.fortum.no/hurtigladere",
              "Comments": null,
              "PhonePrimaryContact": "22 55 54 24",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "chargedrive.no@fortum.com",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 202,
              "Title": "Charge & Drive (Fortum - NO)"
            },
            {
              "WebsiteURL": "https://www.volkswagenleasing.de/de/home/mobilitaetsdienstleistungen/europa_tank_servicekarten/charge-fuel-card.html",
              "Comments": null,
              "PhonePrimaryContact": "+49 531 212-03",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "tankkarten@vwfs.com",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 215,
              "Title": "Charge & Fuel"
            },
            {
              "WebsiteURL": "http://www.chargeyourcar.org.uk/",
              "Comments": null,
              "PhonePrimaryContact": "0191 26 50 500",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "admin@chargeyourcar.org.uk",
              "FaultReportEmail": "admin@chargeyourcar.org.uk",
              "IsRestrictedEdit": false,
              "ID": 20,
              "Title": "Charge Your Car"
            },
            {
              "WebsiteURL": "https://charge.net.nz/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 229,
              "Title": "charge.net.nz"
            },
            {
              "WebsiteURL": "http://www.chargeit-mobility.com/",
              "Comments": null,
              "PhonePrimaryContact": "+49 9321 268 -0700",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@chargeit-mobility.com",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 101,
              "Title": "ChargeIT mobility"
            },
            {
              "WebsiteURL": "http://www.chargelounge.de/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@chargelounge.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 96,
              "Title": "ChargeLounge"
            },
            {
              "WebsiteURL": "http://chargemasterplc.com/",
              "Comments": "Operate networks in both UK and France",
              "PhonePrimaryContact": "+44 (0)1582 400331",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "polarenquiries@chargemasterplc.com",
              "FaultReportEmail": "polarenquiries@chargemasterplc.com",
              "IsRestrictedEdit": false,
              "ID": 8,
              "Title": "Chargemaster"
            },
            {
              "WebsiteURL": "http://chargenet.lk/",
              "Comments": null,
              "PhonePrimaryContact": "+94-1-5551-551 ",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@vega.lk",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 227,
              "Title": "chargeNET (lk)"
            },
            {
              "WebsiteURL": "https://www.chargenow.com/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 173,
              "Title": "ChargeNow"
            },
            {
              "WebsiteURL": "http://chargeplacescotland.org/",
              "Comments": "Operated by ChargeYourCar on behalf of Transport Scotland. CYC cards/app accepted",
              "PhonePrimaryContact": "0141 648 0750",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3315,
              "Title": "Chargeplace Scotland"
            },
            {
              "WebsiteURL": "http://www.chargepoint.net/",
              "Comments": null,
              "PhonePrimaryContact": "1-888-758-4389",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "support@coulombtech.com",
              "FaultReportEmail": "support@coulombtech.com",
              "IsRestrictedEdit": null,
              "ID": 5,
              "Title": "ChargePoint (Coulomb Technologies)"
            },
            {
              "WebsiteURL": "http://chargestar.com.au/",
              "Comments": null,
              "PhonePrimaryContact": "1300 661 895",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3310,
              "Title": "Chargestar (AU)"
            },
            {
              "WebsiteURL": "http://www.lecircuitelectrique.com",
              "Comments": null,
              "PhonePrimaryContact": "1 855 999-8378",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "infocircuitelectrique@caaquebec.com",
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 90,
              "Title": "Circuit Electrique"
            },
            {
              "WebsiteURL": "https://www.clever.dk/",
              "Comments": null,
              "PhonePrimaryContact": "+45 8230 3030",
              "PhoneSecondaryContact": "8230 3030",
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "kundeservice@clever.dk",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 185,
              "Title": "CLEVER "
            },
            {
              "WebsiteURL": null,
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 31,
              "Title": "Clipper Creek"
            },
            {
              "WebsiteURL": "https://www.dewa.gov.ae/",
              "Comments": "Dubai Electricity and Water Authority (DEWA)",
              "PhonePrimaryContact": "+971 4- 601 9999",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3263,
              "Title": "DEWA"
            },
            {
              "WebsiteURL": "http://www.drehstromnetz.de",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 66,
              "Title": "Drehstromnetz"
            },
            {
              "WebsiteURL": "http://driv-eco.com/index.php/en/home-eng/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3316,
              "Title": "Drive Eco"
            },
            {
              "WebsiteURL": "http://mobility.dueenergie.com/",
              "Comments": null,
              "PhonePrimaryContact": "+39 010 27 56 01",
              "PhoneSecondaryContact": "800 93 93 00",
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "mobility@dueenergie.com",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3267,
              "Title": "DUE Energie"
            },
            {
              "WebsiteURL": "http://www.ekobonus.cz/ekologicka-doprava/elektromobilita",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 192,
              "Title": "E.ON (cz)"
            },
            {
              "WebsiteURL": "https://www.eon.dk/privat/strom-til-din-elbil.html",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3251,
              "Title": "E.ON (DK)"
            },
            {
              "WebsiteURL": "https://www.eon.de/de/eonde/pk/produkteUndPreise/E.ON_eMobil/index.htm",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 46,
              "Title": "E.ON (Germany)"
            },
            {
              "WebsiteURL": "https://www.eac.com.cy/En/CustomerService/eCharge/Pages/default.aspx",
              "Comments": "Electricity Authority of Cyprus",
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3294,
              "Title": "EAC e-charge"
            },
            {
              "WebsiteURL": "http://www.eam.de/ueber-uns/unternehmensprofil/standorte/",
              "Comments": null,
              "PhonePrimaryContact": " ",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 140,
              "Title": "EAM"
            },
            {
              "WebsiteURL": "http://www.easy-4-you.ch",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3314,
              "Title": "Easy4You"
            },
            {
              "WebsiteURL": "https://easypark.fi/",
              "Comments": "Parking operator, EV chargers activated by mobile phone app",
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3302,
              "Title": "Easypark"
            },
            {
              "WebsiteURL": "http://www.eaton.com",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 42,
              "Title": "EATON"
            },
            {
              "WebsiteURL": "http://www.ecarni.com/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "ecar@drdni.gov.uk",
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 93,
              "Title": "E-Car "
            },
            {
              "WebsiteURL": "http://e-charge.ro/",
              "Comments": "Renovatio E-Charge",
              "PhonePrimaryContact": "+40 372 756 599",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "office@e-charge.ro",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3271,
              "Title": "e-Charge (Romania)"
            },
            {
              "WebsiteURL": "http://www.echargenet.com",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3309,
              "Title": "echargenet - State Grid Corportation of China"
            },
            {
              "WebsiteURL": "http://www.ecoplug.be/index.html",
              "Comments": null,
              "PhonePrimaryContact": "+32 (0) 483 425 255",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@ecoplug.be",
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 77,
              "Title": "Ecoplug"
            },
            {
              "WebsiteURL": "http://www.ecospazio.it/",
              "Comments": null,
              "PhonePrimaryContact": "+39 0464 401121",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@ecospazio.it",
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 83,
              "Title": "Ecospazio (Italy)"
            },
            {
              "WebsiteURL": "http://www.ecotap.nl/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 179,
              "Title": "Ecotap"
            },
            {
              "WebsiteURL": "http://www.ecotricity.co.uk/for-the-road/",
              "Comments": null,
              "PhonePrimaryContact": "01453 761455",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "fortheroad@ecotricity.co.uk",
              "FaultReportEmail": "fortheroad@ecotricity.co.uk",
              "IsRestrictedEdit": false,
              "ID": 24,
              "Title": "Ecotricity"
            },
            {
              "WebsiteURL": null,
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3312,
              "Title": "EDF"
            },
            {
              "WebsiteURL": "http://www.edp.pt/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3276,
              "Title": "EDP"
            },
            {
              "WebsiteURL": "http://www.edp.pt/",
              "Comments": null,
              "PhonePrimaryContact": " +351 21 001 2500",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 199,
              "Title": "EDP MOP"
            },
            {
              "WebsiteURL": "http://www.edrop.ch/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@edrop.ch",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 208,
              "Title": "eDrop"
            },
            {
              "WebsiteURL": "http://www.ee-mobil.de/",
              "Comments": null,
              "PhonePrimaryContact": "+49 2520 93118 0",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@ee-mobil.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3264,
              "Title": "EE-Mobil"
            },
            {
              "WebsiteURL": "http://www.ewb-duderstadt.de/de/Strom/Ausgezeichnet.html",
              "Comments": null,
              "PhonePrimaryContact": "(05527) 911-0",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@eew-duderstadt.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 171,
              "Title": "EEW Duderstadt"
            },
            {
              "WebsiteURL": "https://www.resaplace.com/",
              "Comments": "Resaplace Parking",
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3281,
              "Title": "Effia"
            },
            {
              "WebsiteURL": "http://www.eins.de/privatkunden/strom/elektromobilitaet/",
              "Comments": null,
              "PhonePrimaryContact": "+49 371 525-5403",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "Robby.Hartl@eins.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 2243,
              "Title": "eins"
            },
            {
              "WebsiteURL": "http://www.e-laad.nl",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 29,
              "Title": "e-Laad"
            },
            {
              "WebsiteURL": "http://electric-driving.colruytgroup.com",
              "Comments": null,
              "PhonePrimaryContact": "+32 (0)2 363 55 45",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "electric-driving@colruytgroup.com",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 232,
              "Title": "electric-driving.colruytgroup.com"
            },
            {
              "WebsiteURL": "http://www.electrodrive-europe.com/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 67,
              "Title": "ElectroDrive"
            },
            {
              "WebsiteURL": "http://www.electrodrive-salzburg.at",
              "Comments": null,
              "PhonePrimaryContact": "0800 / 810 102",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "office@electrodrive-salzburg.at",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 141,
              "Title": "ElectroDrive Salzburg"
            },
            {
              "WebsiteURL": "http://www.mark-e.de/",
              "Comments": null,
              "PhonePrimaryContact": " +49 2331 12322437",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "electrodrive@mark-e.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 184,
              "Title": "ElectroDrive/Mark-E (DE)"
            },
            {
              "WebsiteURL": "http://www.elektro-ljubljana.si/1/Obnovljivi-viri-energije/Polnilna-mesta-za-elektricna-vozila.aspx",
              "Comments": null,
              "PhonePrimaryContact": "+386 1 230 40 00",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@elektro-ljubljana.si",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3258,
              "Title": "Elektro Ljublana"
            },
            {
              "WebsiteURL": "http://www.elektromotive.com",
              "Comments": null,
              "PhonePrimaryContact": "+ 44 (0)1273 704775",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@elektromotive.com",
              "FaultReportEmail": "info@elektromotive.com",
              "IsRestrictedEdit": null,
              "ID": 2,
              "Title": "Elektrobay (UK)"
            },
            {
              "WebsiteURL": "http://www.elektromotive.com",
              "Comments": null,
              "PhonePrimaryContact": "+ 44 (0)1273 704775",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": "http://www.elektromotive.com/html/application_forms.php",
              "ContactEmail": "info@elektromotive.com",
              "FaultReportEmail": "info@elektromotive.com",
              "IsRestrictedEdit": null,
              "ID": 16,
              "Title": "Elektromotive (UK)"
            },
            {
              "WebsiteURL": "http://elen.hep.hr/ELEN-charging-stations.aspx",
              "Comments": null,
              "PhonePrimaryContact": "+385 163 22123",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3311,
              "Title": "Elen"
            },
            {
              "WebsiteURL": "http://ella.at/",
              "Comments": "uses the Be Energised backend",
              "PhonePrimaryContact": "+43 800 203004",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3287,
              "Title": "Ella"
            },
            {
              "WebsiteURL": "http://www.elmo.ee/",
              "Comments": "When calling 651 1911 the price of the call depends on the selected service package. When calling 1911, the price of the call is 0.23 euros (including VAT). The price of content service of calls to call service numbers does not depend on the service package of a customer; it is the same for all callers.",
              "PhonePrimaryContact": "6 511 911",
              "PhoneSecondaryContact": "1911",
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "elmo@kredex.ee",
              "FaultReportEmail": "tugi@elmo.ee",
              "IsRestrictedEdit": null,
              "ID": 76,
              "Title": "ELMO"
            },
            {
              "WebsiteURL": "http://www.e-autozas.hu/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 187,
              "Title": "ELM? "
            },
            {
              "WebsiteURL": "http://www.friedrichshafen.de/wirtschaft-verkehr/emma",
              "Comments": null,
              "PhonePrimaryContact": "+49 7541 603380",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@fn-dienste.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 1238,
              "Title": "emma"
            },
            {
              "WebsiteURL": "http://www.emobitaly.it/come-ricaricare/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3282,
              "Title": "Emobitaly (Italy)"
            },
            {
              "WebsiteURL": "http://www.e-moving.it/",
              "Comments": null,
              "PhonePrimaryContact": "800035151",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "e-moving@a2a.eu",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 82,
              "Title": "E-moving (Italy)"
            },
            {
              "WebsiteURL": "http://www.empora.eu",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 68,
              "Title": "Empora"
            },
            {
              "WebsiteURL": "https://www.enbw.com/privatkunden/tarife-und-produkte/e-mobilitaet/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 86,
              "Title": "EnBW (D)"
            },
            {
              "WebsiteURL": "http://www.endesa.es/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 207,
              "Title": "Endesa"
            },
            {
              "WebsiteURL": "https://www.eneldrive.it/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 80,
              "Title": "Enel (Italy)"
            },
            {
              "WebsiteURL": "https://www.enercity.de/privatkunden/mobilitaet/e-mobilitaet/formular-reg-e-tanken/index.jsx",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 121,
              "Title": "EnerCity"
            },
            {
              "WebsiteURL": "https://www.ene-eifel.de/privatkunden-strom/e-mobilitt",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "service@ene-eifel.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3260,
              "Title": "Energie der Eifel"
            },
            {
              "WebsiteURL": "http://energiedienst.de",
              "Comments": null,
              "PhonePrimaryContact": "+49 7623 921818",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "kundenservice@energiedienst.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 224,
              "Title": "EnergieDienst"
            },
            {
              "WebsiteURL": "http://ebt-halblech.de/",
              "Comments": null,
              "PhonePrimaryContact": "+49 8368 9280",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "kontakt@ebt-halblech.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 219,
              "Title": "Energieversorgung Buching-Trauchgau"
            },
            {
              "WebsiteURL": "http://www.enewa.de/",
              "Comments": null,
              "PhonePrimaryContact": "0228 / 377368 0",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@enewa.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 172,
              "Title": "ENEWA"
            },
            {
              "WebsiteURL": "http://www.engenie.co.uk/",
              "Comments": null,
              "PhonePrimaryContact": "03300 1000 50",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@engenie.co.uk",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 203,
              "Title": "Engenie"
            },
            {
              "WebsiteURL": "http://www.enovates.com/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 181,
              "Title": "eNovates"
            },
            {
              "WebsiteURL": "http://www.enovos.lu/particuliers/ecomobilite",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 52,
              "Title": "Enovos"
            },
            {
              "WebsiteURL": "http://www.enspirion.pl/?page_id=728",
              "Comments": null,
              "PhonePrimaryContact": "+48 58 888 88 00",
              "PhoneSecondaryContact": "+48 785 888 805",
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "biuro@enspirion.pl",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 216,
              "Title": "Enspirion"
            },
            {
              "WebsiteURL": "https://www.entega.de/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3297,
              "Title": "Entega"
            },
            {
              "WebsiteURL": "https://www.eocharging.com/",
              "Comments": null,
              "PhonePrimaryContact": "+44 333 772 0383 ",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3298,
              "Title": "EO Charging"
            },
            {
              "WebsiteURL": "http://esarj.com/uyelik",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 70,
              "Title": "E?arj (TR)"
            },
            {
              "WebsiteURL": "http://www.esb.ie/electric-cars/index.jsp",
              "Comments": null,
              "PhonePrimaryContact": " 1890 372 387",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "ecars@esb.ie",
              "FaultReportEmail": "ecars@esb.ie",
              "IsRestrictedEdit": null,
              "ID": 22,
              "Title": "ESB Ecars"
            },
            {
              "WebsiteURL": "http://www.essent.nl/content/particulier/producten/elektrisch_rijden/index.html",
              "Comments": null,
              "PhonePrimaryContact": "0800 377 36 83",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "elektrischrijden@essent.nl",
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 78,
              "Title": "Essent (NL)"
            },
            {
              "WebsiteURL": "http://www.e-mobilitat.cat/",
              "Comments": null,
              "PhonePrimaryContact": "+34 900 250 260",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 209,
              "Title": "Estabanell Energia"
            },
            {
              "WebsiteURL": "http://www.estonteco.eu/live/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 205,
              "Title": "Estonteca"
            },
            {
              "WebsiteURL": "http://etop.sk/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 222,
              "Title": "ETOP"
            },
            {
              "WebsiteURL": "http://www.e-totem.fr/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 55,
              "Title": "e-totem"
            },
            {
              "WebsiteURL": "http://www.evbility.eu/",
              "Comments": null,
              "PhonePrimaryContact": "+390396015174",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@evbility.com",
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 87,
              "Title": "EVbility (Italy)"
            },
            {
              "WebsiteURL": "http://www.ev-box.com",
              "Comments": null,
              "PhonePrimaryContact": "+31 (0)88 77 55 444",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@ev-box.com",
              "FaultReportEmail": "info@ev-box.com",
              "IsRestrictedEdit": null,
              "ID": 73,
              "Title": "EV-Box"
            },
            {
              "WebsiteURL": "https://evcharge.online/",
              "Comments": "Rolec",
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3295,
              "Title": "evcharge.online"
            },
            {
              "WebsiteURL": "http://www.evd-dormagen.de/",
              "Comments": null,
              "PhonePrimaryContact": "+49 2133 97150",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3254,
              "Title": "Evd Dormagen"
            },
            {
              "WebsiteURL": "https://www.evgo.com/",
              "Comments": "Imported by ADFC import",
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 15,
              "Title": "eVgo Network"
            },
            {
              "WebsiteURL": "http://www.swiss-emobility.ch/home/evite.html",
              "Comments": null,
              "PhonePrimaryContact": "+41 (0)58 827 34 09 ",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@swiss-emobility.ch",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 110,
              "Title": "EVite (ch)"
            },
            {
              "WebsiteURL": "http://www.evmapa.cz",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 221,
              "Title": "Evmapa (CZ)"
            },
            {
              "WebsiteURL": "http://www.evnet.nl/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 178,
              "Title": "EVnetNL"
            },
            {
              "WebsiteURL": "https://www.evpass.ch/",
              "Comments": "Android & IOS Apps, Swisspass travel RFID can be registered",
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3272,
              "Title": "EVPass (CH)"
            },
            {
              "WebsiteURL": "http://www.ev-point.be",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 49,
              "Title": "EV-Point"
            },
            {
              "WebsiteURL": "https://www.energieversorgung-sylt.de/mobilitaet/e-mobilitaet/",
              "Comments": null,
              "PhonePrimaryContact": "04651 925-925",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "kundenservice@energieversorgung-sylt.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 137,
              "Title": "EVS Energieversorgung Sylt "
            },
            {
              "WebsiteURL": "",
              "Comments": "Imported by ADFC import",
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 11,
              "Title": "EVSE LLC WebNet"
            },
            {
              "WebsiteURL": "https://evway.net",
              "Comments": null,
              "PhonePrimaryContact": "+39 02 35954219",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3291,
              "Title": "Evway"
            },
            {
              "WebsiteURL": "http://shop.e-wald.eu/produkt/e-wald-ladekarte-monatskarte/",
              "Comments": null,
              "PhonePrimaryContact": "+49 9923 8045-310",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@e-wald.eu",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 189,
              "Title": "E-Wald"
            },
            {
              "WebsiteURL": "http://www.stadtwerke-bruchsal.de/",
              "Comments": null,
              "PhonePrimaryContact": "+49 7251 7060",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@stadtwerke-bruchsal.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 2241,
              "Title": "EWB"
            },
            {
              "WebsiteURL": "http://www.ewe.de/privatkunden/ewe-stromtankstellen.php",
              "Comments": null,
              "PhonePrimaryContact": "04418030",
              "PhoneSecondaryContact": "08001014432",
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "stromtankkarte@ewe.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 127,
              "Title": "EWE"
            },
            {
              "WebsiteURL": "http://www.ewi-isernhagen.de/energieeffizienz/elektromobilitaet.aspx",
              "Comments": null,
              "PhonePrimaryContact": " 0511 6165475",
              "PhoneSecondaryContact": " 0511 6165476",
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "kontakt@ewi-isernhagen.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 145,
              "Title": "EWI  Energiewerke Isernhagen"
            },
            {
              "WebsiteURL": "http://www.ewr-e-mobil.de/",
              "Comments": null,
              "PhonePrimaryContact": "+49 6241 848-0",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@ewr-e-mobil.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 109,
              "Title": "EWR gmbh e-mobile"
            },
            {
              "WebsiteURL": "http://www.fastned.nl/",
              "Comments": null,
              "PhonePrimaryContact": "+31 (0)20-7155316",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "contact@fastned.nl",
              "FaultReportEmail": "contact@fastned.nl",
              "IsRestrictedEdit": null,
              "ID": 74,
              "Title": "FastNed"
            },
            {
              "WebsiteURL": "http://recarga.fenieenergia.es/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3279,
              "Title": "Fenie Energa (Spain)"
            },
            {
              "WebsiteURL": "https://flo.ca/",
              "Comments": "RseauVER / VERnetwork        http://www.reseauver.com/          (877) 505-2674          service@reseauver.com",
              "PhonePrimaryContact": "+1 877 505-2674",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@flo.ca",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 89,
              "Title": "flo"
            },
            {
              "WebsiteURL": "http://www.flowcharging.com/",
              "Comments": null,
              "PhonePrimaryContact": "+31 4012734",
              "PhoneSecondaryContact": "+31 4012735",
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@flow-nederland.nl",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 197,
              "Title": "FLOW Charging"
            },
            {
              "WebsiteURL": "http://www.fortisis.eu",
              "Comments": "Operator based on Greence and Cyprus",
              "PhonePrimaryContact": "+30.215.54.09.814 (Greece)",
              "PhoneSecondaryContact": "+357.22.31.63.18 (Cyprus)",
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@fortisis.eu",
              "FaultReportEmail": "info@fortisis.eu",
              "IsRestrictedEdit": false,
              "ID": 85,
              "Title": "FORTISIS"
            },
            {
              "WebsiteURL": "http://www.g2mobility.com/",
              "Comments": null,
              "PhonePrimaryContact": "01 45 34 25 34.",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 116,
              "Title": "G2mobility"
            },
            {
              "WebsiteURL": "http://galactico.pl/",
              "Comments": null,
              "PhonePrimaryContact": " +58 52 44 500",
              "PhoneSecondaryContact": " +48 68 328 20 89",
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "kontakt@galactico.pl",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 218,
              "Title": "Galactico.pl"
            },
            {
              "WebsiteURL": "http://www.gardauno.it/la-mappa-dei-punti-di-ricarica/",
              "Comments": null,
              "PhonePrimaryContact": "800-133 966",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 204,
              "Title": "GardaUno"
            },
            {
              "WebsiteURL": "https://www.gewattstation.com/connect/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 1242,
              "Title": "GE WattStation"
            },
            {
              "WebsiteURL": "http://www.ggew.de/UN/Elektromobilitaet",
              "Comments": null,
              "PhonePrimaryContact": "06251 1301-0",
              "PhoneSecondaryContact": "0800 80 30 300",
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@ggew.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 163,
              "Title": "GGEW"
            },
            {
              "WebsiteURL": "http://www.recargagic.com/",
              "Comments": null,
              "PhonePrimaryContact": "(+34) 902 103 498",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@recargagic.com",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 210,
              "Title": "GIC"
            },
            {
              "WebsiteURL": "http://gofastcharge.com/",
              "Comments": "Accepts Hubject roaming partners of Swisscharge",
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@gofastcharge.com",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3313,
              "Title": "GOFAST (Gotthard Fastcharge)"
            },
            {
              "WebsiteURL": "http://green-frontiers.net/",
              "Comments": null,
              "PhonePrimaryContact": " +94 11 2884738",
              "PhoneSecondaryContact": "+94710710271",
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@green-frontiers.net",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 226,
              "Title": "Green Frontiers (lk)"
            },
            {
              "WebsiteURL": "http://greenlandmobility.it/",
              "Comments": null,
              "PhonePrimaryContact": "02/27208182",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "redazione@greenlandmobility.it",
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 84,
              "Title": "Green Land Mobility (Italy)"
            },
            {
              "WebsiteURL": "http://www.greenflux.nl/",
              "Comments": null,
              "PhonePrimaryContact": "088 60 50 700",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@greenflux.nl",
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 63,
              "Title": "Greenflux"
            },
            {
              "WebsiteURL": "http://www.greenlots.com",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 59,
              "Title": "GreenLots"
            },
            {
              "WebsiteURL": "http://www.plug-in.sk/",
              "Comments": null,
              "PhonePrimaryContact": "+421 911 668 770",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@greenway.sk",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 88,
              "Title": "Greenway (sk)"
            },
            {
              "WebsiteURL": "http://gronnkontakt.no/",
              "Comments": null,
              "PhonePrimaryContact": "47 67 08 00 ",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3247,
              "Title": "Grnn Kontakt"
            },
            {
              "WebsiteURL": "http://www.harzenergie.de/index.cfm?fuseaction=portal.showcontent&viewmode=content&num_obj_id=30312&language=de&menu=30305&rootmenu=25325&page=30306&bp=30306",
              "Comments": null,
              "PhonePrimaryContact": "05522/503-9345",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "mobilitaet@harzenergie.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 122,
              "Title": "Harz Energie"
            },
            {
              "WebsiteURL": "http://www.helen.fi/",
              "Comments": null,
              "PhonePrimaryContact": " 010 802 802 ",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 95,
              "Title": "Helsingin Energia"
            },
            {
              "WebsiteURL": "http://www.hkevpower.com/",
              "Comments": null,
              "PhonePrimaryContact": "+852 2210 7122",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "contact@hkevpower.com",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 111,
              "Title": "Hong Kong EV Power"
            },
            {
              "WebsiteURL": "https://www.iberdrola.es/clientes/hogar/movilidad-verde/recarga",
              "Comments": null,
              "PhonePrimaryContact": "900 22 45 22",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 2247,
              "Title": "Iberdrola"
            },
            {
              "WebsiteURL": "https://www.ibil.es",
              "Comments": null,
              "PhonePrimaryContact": "902 540 810",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "ibil@ibil.es",
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 91,
              "Title": "IBIL (Es)"
            },
            {
              "WebsiteURL": "http://inchanet.cz/products_en.html#fast_charging_station",
              "Comments": null,
              "PhonePrimaryContact": "+420 776 333 155",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@inchanet.cz",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 191,
              "Title": "inChaNet"
            },
            {
              "WebsiteURL": "https://www.rwe-mobility.com/",
              "Comments": null,
              "PhonePrimaryContact": "+49(0) 800 88 88 862",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 105,
              "Title": "Innogy SE (RWE eMobility)"
            },
            {
              "WebsiteURL": "http://instavolt.co.uk/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3296,
              "Title": "InstaVolt Ltd"
            },
            {
              "WebsiteURL": "http://www.ionity.eu/",
              "Comments": "Pan-European High-power CCS charging network joint venture of BMW Group, Daimler AG, Ford Motor Company and the Volkswagen Group",
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3299,
              "Title": "Ionity"
            },
            {
              "WebsiteURL": "https://www.jec.co.uk/your-home/electric-vehicles/charging-your-ev/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3288,
              "Title": "Jersey Electricity Evolve"
            },
            {
              "WebsiteURL": "http://www.juicepoint.co.nz/",
              "Comments": null,
              "PhonePrimaryContact": "+64 9 354 3869",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@juicepoint.co.nz",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 120,
              "Title": "JuicePoint"
            },
            {
              "WebsiteURL": "https://www.kelag.at/privat/kelag-autostrom-111.htm",
              "Comments": null,
              "PhonePrimaryContact": "+43 463 525 9660",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3285,
              "Title": "Kelag AG"
            },
            {
              "WebsiteURL": "http://www.kiwhipass.fr/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "contact@kiwhipass.fr",
              "FaultReportEmail": "assistance@kiwhipass.fr",
              "IsRestrictedEdit": null,
              "ID": 62,
              "Title": "KiWhi Pass"
            },
            {
              "WebsiteURL": "https://ladefoxx.de/",
              "Comments": null,
              "PhonePrimaryContact": "05261 255-251",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@ladefoxx.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 106,
              "Title": "Ladefoxx"
            },
            {
              "WebsiteURL": "http://ladenetz.de",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": "info@smartlab-gmbh.de",
              "IsRestrictedEdit": false,
              "ID": 69,
              "Title": "ladenetz.de"
            },
            {
              "WebsiteURL": "http://www.latvenergo.lv/eng/tenders_and_offers/charging_points_for_electric_vehicles_/",
              "Comments": null,
              "PhonePrimaryContact": "+371 67 723 511",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@latvenergo.lv",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 182,
              "Title": "Latvenergo"
            },
            {
              "WebsiteURL": "http://lenenergo.ru/ev/",
              "Comments": "Utility in St. Petersburg, Russia",
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3286,
              "Title": "Lenenergo"
            },
            {
              "WebsiteURL": null,
              "Comments": "Belgian charging network",
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 38,
              "Title": "Lidl"
            },
            {
              "WebsiteURL": "http://w41.bcn.cat/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 206,
              "Title": "LIVE Barcelona"
            },
            {
              "WebsiteURL": "http://www.elektrofahrzeuge.lsw.de/",
              "Comments": null,
              "PhonePrimaryContact": " 05361 1893600",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "mobilitaet@lsw.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 128,
              "Title": "LSW Energie"
            },
            {
              "WebsiteURL": null,
              "Comments": "Obsolete inductive paddle charging standard",
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 41,
              "Title": "MAGNE Charge"
            },
            {
              "WebsiteURL": "https://www.mainova.de/privatkunden/mobilitaet/stromtankstellen.html",
              "Comments": null,
              "PhonePrimaryContact": "069 / 213-24221 ",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "Efahrzeug-Foerderung@mainova.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 165,
              "Title": "Mainova"
            },
            {
              "WebsiteURL": "http://www.manxelectricity.com",
              "Comments": null,
              "PhonePrimaryContact": "(01624) 687687",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "mea@gov.im",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 135,
              "Title": "Manx Electricity Authority"
            },
            {
              "WebsiteURL": "http://www.caib.es/sites/energiaicanviclimatic/ca/mobilitat_elactrica_a_les_illes_balears_melib/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3283,
              "Title": "MELIB (ES)"
            },
            {
              "WebsiteURL": "http://thefastchargernetwork.com/",
              "Comments": null,
              "PhonePrimaryContact": "+ 31 20 7719 026",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "contact@mistergreen.nl",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 112,
              "Title": "MisterGreen, The Fast Charger Network"
            },
            {
              "WebsiteURL": "http://www.mobecpoint.com",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@mobecpoint.com",
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 92,
              "Title": "MobecPoint (Es)"
            },
            {
              "WebsiteURL": "http://www.mobib.be/index.htm",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 79,
              "Title": "Mobib (Belgium) "
            },
            {
              "WebsiteURL": "http://www.mobie.pt",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 21,
              "Title": "Mobie.pt"
            },
            {
              "WebsiteURL": "http://www.move-net.ch/de/",
              "Comments": null,
              "PhonePrimaryContact": "0800 29 29 29",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3274,
              "Title": "MOVE (CH)"
            },
            {
              "WebsiteURL": "https://www.bayernwerk.de/cps/rde/xchg/bayernwerk/hs.xsl/278.htm",
              "Comments": null,
              "PhonePrimaryContact": "+49 89 61413-536",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "BAG-e-mobility@bayernwerk.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 201,
              "Title": "Mnchen Umland"
            },
            {
              "WebsiteURL": "https://www.mvmpartner.hu/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "mvmp@mvmp.hu",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3289,
              "Title": "MVM Partner Zrt."
            },
            {
              "WebsiteURL": "https://www.n-ergie.de/header/die-n-ergie/aktiv-fuer-die-umwelt/elektromobilitaet.html",
              "Comments": null,
              "PhonePrimaryContact": "+49 911 80201",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "elektromobilitaet@n-ergie.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 234,
              "Title": "N-ERGIE"
            },
            {
              "WebsiteURL": "http://www.nissan.de/DE/de/vehicle/electric-vehicles/leaf/charging-and-battery/freistrom-info.html",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 167,
              "Title": "Nissan DE Freistrom"
            },
            {
              "WebsiteURL": "http://www.nissan.co.uk",
              "Comments": "Network of franchised Nissan Dealers",
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 7,
              "Title": "Nissan UK Dealer Network"
            },
            {
              "WebsiteURL": null,
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 6,
              "Title": "Nissan US Dealer Network"
            },
            {
              "WebsiteURL": "http://www.mobiliti.hu/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3307,
              "Title": "NKM mobiliti"
            },
            {
              "WebsiteURL": "http://www.nomadpower.eu",
              "Comments": null,
              "PhonePrimaryContact": "+31 348435512",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@nomadpower.eu",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 196,
              "Title": "Nomadpower"
            },
            {
              "WebsiteURL": "https://www.nrgevgo.com/",
              "Comments": null,
              "PhonePrimaryContact": "855-509-5581",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3252,
              "Title": "NRG EVgo"
            },
            {
              "WebsiteURL": "http://www.nuon.com/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 177,
              "Title": "Nuon"
            },
            {
              "WebsiteURL": "https://www.opconnect.com/",
              "Comments": null,
              "PhonePrimaryContact": "1-855-885-9571",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "nisaacs@opconnect.com",
              "FaultReportEmail": "support@opconnect.com",
              "IsRestrictedEdit": false,
              "ID": 107,
              "Title": "OpConnect"
            },
            {
              "WebsiteURL": "http://www.ores.net/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 1236,
              "Title": "ORES"
            },
            {
              "WebsiteURL": "http://www.on.is/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "on@on.is",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 102,
              "Title": "Orka Nttrunnar"
            },
            {
              "WebsiteURL": "https://www.osterholzer-stadtwerke.de/service/fahren-mit-strom-erdgas/strom/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 124,
              "Title": "Osterholzer Stadtwerke"
            },
            {
              "WebsiteURL": "http://www.ovag-energie.de/oe/ovag-energie.nsf/c/Umwelt,E-Mobilit%C3%A4t",
              "Comments": null,
              "PhonePrimaryContact": "+49 6031 6848-0",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "service@ovag-energie.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 188,
              "Title": "OVAG Energie"
            },
            {
              "WebsiteURL": "http://www.park-charge.ch/",
              "Comments": null,
              "PhonePrimaryContact": "+41 44 820 24 55  ",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "park-charge@gmx.net",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 97,
              "Title": "Park & Charge (CH)"
            },
            {
              "WebsiteURL": "http://www.park-charge.de/",
              "Comments": null,
              "PhonePrimaryContact": "030-32 59 91 80 ",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@park-charge.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 98,
              "Title": "Park & Charge (D)"
            },
            {
              "WebsiteURL": "http://www.petrol.eu/road/car/electrical-mobility-petrol",
              "Comments": "PETROL is a fuel supplier and runs charge points in Slovenia, part of the CENTRAL EUROPEAN GREEN CORRIDORS (CEGC) PROJECT",
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 13,
              "Title": "PETROL"
            },
            {
              "WebsiteURL": "http://www.pfalzwerke.de/",
              "Comments": null,
              "PhonePrimaryContact": "+49 6215850",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@pfalzwerke.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 1240,
              "Title": "Pfalzwerke"
            },
            {
              "WebsiteURL": "https://plugnroll.com/elektroautofahrer/netzwerk/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3280,
              "Title": "Plug n Roll"
            },
            {
              "WebsiteURL": "http://www.pluggedinmidlands.co.uk",
              "Comments": null,
              "PhonePrimaryContact": "0845 838 0551",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "enquiries@pluggedinmidlands.co.uk",
              "FaultReportEmail": "enquiries@pluggedinmidlands.co.uk",
              "IsRestrictedEdit": false,
              "ID": 71,
              "Title": "Plugged In Midlands (UK)"
            },
            {
              "WebsiteURL": "http://www.pluginncw.com/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "plugincenter@gmail.com",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 119,
              "Title": "Plug-In North Central Washington"
            },
            {
              "WebsiteURL": "https://www.plugsurfing.com/",
              "Comments": null,
              "PhonePrimaryContact": "+49 30 9599814 10 ",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "support@plugsurfing.com",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3266,
              "Title": "PlugSurfing"
            },
            {
              "WebsiteURL": "http://www.pod-point.com/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": "020 7247 4114",
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "enquiries@pod-point.com",
              "FaultReportEmail": "enquiries@pod-point.com",
              "IsRestrictedEdit": null,
              "ID": 3,
              "Title": "POD Point (UK)"
            },
            {
              "WebsiteURL": "http://www.polarnetwork.com/",
              "Comments": "Owned by Chargemaster in the UK",
              "PhonePrimaryContact": "01582 400331",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "polarenquiries@chargemasterplc.com",
              "FaultReportEmail": "polarenquiries@chargemasterplc.com",
              "IsRestrictedEdit": false,
              "ID": 32,
              "Title": "POLAR (UK)"
            },
            {
              "WebsiteURL": "http://polyfazer.cz/",
              "Comments": null,
              "PhonePrimaryContact": "+420 270 007 900?",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@polyfazer.cz",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 1239,
              "Title": "Polyfazer"
            },
            {
              "WebsiteURL": "http://power-station.be/power-card-laadpas/",
              "Comments": null,
              "PhonePrimaryContact": "+32 (0)488 985 200",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@power-station.be",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 114,
              "Title": "Power Station"
            },
            {
              "WebsiteURL": "https://www.pre.cz/cs/profil-spolecnosti/dalsi-aktivity-pre/premobilita/",
              "Comments": null,
              "PhonePrimaryContact": "840 550 055",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 190,
              "Title": "PRE (cz)"
            },
            {
              "WebsiteURL": "http://www.prioenergy.com/produtos-e-servicos/mobilidade-electrica/",
              "Comments": "Prio.E Mobility Solutions",
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 200,
              "Title": "PRIOE"
            },
            {
              "WebsiteURL": "http://chargestar.com.au/rac-electric-highway/",
              "Comments": null,
              "PhonePrimaryContact": "1300 661895",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@chargestar.com.au",
              "FaultReportEmail": "info@chargestar.com.au",
              "IsRestrictedEdit": false,
              "ID": 194,
              "Title": "RAC Electric Highway/ChargeStar"
            },
            {
              "WebsiteURL": "http://www.raumanenergia.fi/yritys/fi_FI/sahkoautoilu/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 133,
              "Title": "Rauman Energia"
            },
            {
              "WebsiteURL": "https://revitalizechargingsolutions.com/",
              "Comments": null,
              "PhonePrimaryContact": " 1-817-659-1030",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@revitalizechargingsolutions.com",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 217,
              "Title": "RCS (Revitalize Charging Solutions)"
            },
            {
              "WebsiteURL": "http://www.recharge-power.com/",
              "Comments": "Imported by ADFC import",
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 12,
              "Title": "RechargeAccess"
            },
            {
              "WebsiteURL": "http://www.renault.fr/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3275,
              "Title": "Renault"
            },
            {
              "WebsiteURL": "https://www.reveocharge.com/",
              "Comments": "Sud de la France",
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3257,
              "Title": "Rvo (FR)"
            },
            {
              "WebsiteURL": "https://www.revoltaegypt.com/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@revoltaegypt.com",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3293,
              "Title": "Revolta Egypt"
            },
            {
              "WebsiteURL": "https://www.rewag.de/privatkunden/strom/rewariostrommobil.html",
              "Comments": null,
              "PhonePrimaryContact": " +49 941 6010",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@rewag.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3265,
              "Title": "REWAG"
            },
            {
              "WebsiteURL": "http://www.rheinenergie.com/TankEn-Registrierung",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "energieberatung@rheinenergie.com",
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 72,
              "Title": "RheinEnergie AG"
            },
            {
              "WebsiteURL": "https://re-fd.de/",
              "Comments": null,
              "PhonePrimaryContact": "+49 661 12-100",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@re-fd.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 2242,
              "Title": "RhnEnergie"
            },
            {
              "WebsiteURL": "http://rpsnet.cz",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3250,
              "Title": "RPSnet"
            },
            {
              "WebsiteURL": "http://www.npower.com/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 104,
              "Title": "RWE / Npower"
            },
            {
              "WebsiteURL": "http://www.essent.nl",
              "Comments": "Essent is part of the RWE group",
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 48,
              "Title": "RWE Mobility/Essent"
            },
            {
              "WebsiteURL": "http://www.sahkoinenliikenne.fi/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@sahkoinenliikenne.fi",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 113,
              "Title": "Shkinenliikenne (fi)"
            },
            {
              "WebsiteURL": "https://www.schleswiger-stadtwerke.de/content/unternehmen/emobilitaet/",
              "Comments": null,
              "PhonePrimaryContact": "(0 46 21) 801-414",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "matthias.beier@schleswiger-stadtwerke.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 139,
              "Title": "Schleswiger Stadtwerke"
            },
            {
              "WebsiteURL": "http://www.schnell-laden-berlin.de",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "wirkuemmernuns@schnell-laden-berlin.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 223,
              "Title": "Schnell Laden Berlin"
            },
            {
              "WebsiteURL": "https://www.swu.de/privatkunden/energie-wasser/elektromobilitaet/kunde-werden.html",
              "Comments": null,
              "PhonePrimaryContact": "+49 731 166-8810",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@swu.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 2245,
              "Title": "Schwabencard"
            },
            {
              "WebsiteURL": "http://www.sde76.fr/",
              "Comments": "Can be Accessed via Sodetrel pass",
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3262,
              "Title": "SDE76 (FR)"
            },
            {
              "WebsiteURL": "http://sdey.fr/nos-missions/mobilite-electrique/",
              "Comments": "Syndicat Dpartemental d'nergies de l'Yonne",
              "PhonePrimaryContact": "03 86 52 22 00",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 33,
              "Title": "SDEY (Fr)"
            },
            {
              "WebsiteURL": "http://www.semacharge.com",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 39,
              "Title": "SemaCharge Network"
            },
            {
              "WebsiteURL": "http://www.shorepower.com/",
              "Comments": "Imported by ADFC import",
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 14,
              "Title": "Shorepower"
            },
            {
              "WebsiteURL": null,
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3308,
              "Title": "SIEL"
            },
            {
              "WebsiteURL": "http://www.silfi.it/IT/index.php?id=63&label=Mappa%20colonnine%20ricarica%20veicoli%20elettrici%22",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 81,
              "Title": "Silfi (Italy)"
            },
            {
              "WebsiteURL": "https://sepredaj.seas.sk/elektricka-mobilita",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@sepredaj.sk",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 2246,
              "Title": "Slovensk elektrrne"
            },
            {
              "WebsiteURL": "https://smatrics.com/ladenetz#",
              "Comments": "SMATRICS runs charge points in Austria and Germany, part of the CENTRAL EUROPEAN GREEN CORRIDORS (CEGC) PROJECT",
              "PhonePrimaryContact": " +43 1 532 24 00",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@smatrics.com",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3253,
              "Title": "SMATRICS Netz"
            },
            {
              "WebsiteURL": "http://www.sodetrel-mobilite.fr/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "contact@sodetrel.fr",
              "FaultReportEmail": "sav@sodetrel.fr",
              "IsRestrictedEdit": false,
              "ID": 213,
              "Title": "Sodetrel"
            },
            {
              "WebsiteURL": "https://en.sodo.si/fast-charging-stations/list-of-active-fast-charging-stations-on-slovenian-highways",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3255,
              "Title": "SODO"
            },
            {
              "WebsiteURL": "http://www.sofos.es/productos-y-servicios/estaciones-de-recarga/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 211,
              "Title": "sofos"
            },
            {
              "WebsiteURL": "http://www.sourceeast.net/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "sourceeast@sourceeast.net",
              "FaultReportEmail": "sourceeast@sourceeast.net",
              "IsRestrictedEdit": false,
              "ID": 60,
              "Title": "Source East (UK)"
            },
            {
              "WebsiteURL": "http://www.sourcelondon.net/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": "https://www.sourcelondon.net/membership",
              "ContactEmail": "membership@sourcelondon.net",
              "FaultReportEmail": "membership@sourcelondon.net",
              "IsRestrictedEdit": null,
              "ID": 25,
              "Title": "Source London"
            },
            {
              "WebsiteURL": "http://www.sourcewest.info/",
              "Comments": null,
              "PhonePrimaryContact": "0191 265 0500",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@sourcewest.info",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 118,
              "Title": "Source West"
            },
            {
              "WebsiteURL": "http://sparkev.lk/smart-plug-in-membership/",
              "Comments": null,
              "PhonePrimaryContact": "+94117590590",
              "PhoneSecondaryContact": "+94117727539",
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@sparkev.lk",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 228,
              "Title": "Spark EV (lk)"
            },
            {
              "WebsiteURL": "http://www.sw-magdeburg.de/",
              "Comments": null,
              "PhonePrimaryContact": " 0800 0796796",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@sw-magdeburg.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 152,
              "Title": "Stdtische Werke Magdeburg"
            },
            {
              "WebsiteURL": "http://www.stadtwerke-clausthal.de/",
              "Comments": null,
              "PhonePrimaryContact": " 05323 715-0 ",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@stadtwerke-clausthal.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 162,
              "Title": "Stadtwerke Clausthal-Zellerfeld"
            },
            {
              "WebsiteURL": "https://www.dvv-dessau.de/",
              "Comments": null,
              "PhonePrimaryContact": " 0340 8992000",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 158,
              "Title": "Stadtwerke Dessau"
            },
            {
              "WebsiteURL": "https://www.swd-ag.de",
              "Comments": null,
              "PhonePrimaryContact": "0211/821-8210",
              "PhoneSecondaryContact": "0211/821-4093 (Emergency Contact Number 24/7)",
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "elektromobilitaet@swd-ag.de",
              "FaultReportEmail": "elektromobilitaet@swd-ag.de",
              "IsRestrictedEdit": false,
              "ID": 235,
              "Title": "Stadtwerke Dsseldorf AG"
            },
            {
              "WebsiteURL": "http://www.stadtwerke-elmshorn.de/cms/Dienstleistungen/E-Mobilitaet/Stadtwerke-investieren-in-E-Mobilitaet-.html",
              "Comments": null,
              "PhonePrimaryContact": "(04121) 645-0",
              "PhoneSecondaryContact": "(04121) 645-751",
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@stadtwerke-elmshorn.de",
              "FaultReportEmail": "service.technik@stadtwerke-elmshorn.de",
              "IsRestrictedEdit": false,
              "ID": 170,
              "Title": "Stadtwerke Elmshorn"
            },
            {
              "WebsiteURL": "http://www.swgoe.de/",
              "Comments": null,
              "PhonePrimaryContact": "0551 301290",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 132,
              "Title": "Stadtwerke Gttingen"
            },
            {
              "WebsiteURL": "http://www.stadtwerke-gvm.de/",
              "Comments": null,
              "PhonePrimaryContact": "03881 78450",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@stadtwerke-gvm.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 151,
              "Title": "Stadtwerke Grevesmhlen"
            },
            {
              "WebsiteURL": "http://www.swhdl.de/strom/e_mobility/",
              "Comments": null,
              "PhonePrimaryContact": "03904 725995 ",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 148,
              "Title": "Stadtwerke Haldensleben  SWH"
            },
            {
              "WebsiteURL": "http://www.evh.de/EVH/Privatkunden/Natuerlich-EVH/Mit-Strom-fahren/",
              "Comments": null,
              "PhonePrimaryContact": "(0345) 581 33 33",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "kontakt@evh.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 161,
              "Title": "Stadtwerke Halle"
            },
            {
              "WebsiteURL": "https://www.stadtwerke-hameln.de/service/beratung.html",
              "Comments": null,
              "PhonePrimaryContact": " 05151-7880",
              "PhoneSecondaryContact": " 0800 788 0000",
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 146,
              "Title": "Stadtwerke Hameln"
            },
            {
              "WebsiteURL": "http://www.swl.de/web/swl/DE/Unternehmen/Elektromobilitaet/Elektromobilitaet.htm",
              "Comments": null,
              "PhonePrimaryContact": "0341 121-6404",
              "PhoneSecondaryContact": "0800 121-3001 ",
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "elektromobilitaet@swl.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 155,
              "Title": "Stadtwerke Leipzig SWL"
            },
            {
              "WebsiteURL": "http://www.swhl.de/e-mobilitaet/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 123,
              "Title": "Stadtwerke Lbeck"
            },
            {
              "WebsiteURL": "https://www.stadtwerke-muenster.de/privatkunden/strom/alle-stromprodukte/e-mobilitaet/uebersicht.html",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 225,
              "Title": "Stadtwerke Mnster"
            },
            {
              "WebsiteURL": "http://www.stadtwerke-neumuenster.de/wDeutsch/privatkunden/unser-einsatz-fuer-die-umwelt/e-mobilitaet.php?navid=172",
              "Comments": null,
              "PhonePrimaryContact": " 04321 202654 ",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "swn@swn.net",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 147,
              "Title": "Stadtwerke Neumnster (SWN)."
            },
            {
              "WebsiteURL": "http://www.stadtwerke-northeim.de/site/de/580/stromtankstelle.html",
              "Comments": null,
              "PhonePrimaryContact": "(0 55 51) 60 05 - 0 ",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@stadtwerke-northeim.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 131,
              "Title": "Stadtwerke Northeim  SWN"
            },
            {
              "WebsiteURL": "http://www.stadtwerke-rinteln.de/",
              "Comments": null,
              "PhonePrimaryContact": "05751 700-0 ",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@stadtwerke-rinteln.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 143,
              "Title": "Stadtwerke Rinteln"
            },
            {
              "WebsiteURL": "http://www.swrag.de/",
              "Comments": null,
              "PhonePrimaryContact": " 0381 8051901",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "kundenzentrum@swrag.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 154,
              "Title": "Stadtwerke Rostock"
            },
            {
              "WebsiteURL": "http://www.stadtwerke-soest.de/index.php?id=280",
              "Comments": null,
              "PhonePrimaryContact": "02921.392-0",
              "PhoneSecondaryContact": "02921.392-150",
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@stadtwerke-soest.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 149,
              "Title": "Stadtwerke Soest"
            },
            {
              "WebsiteURL": "http://www.stadtwerke-stralsund.de/energie/elektrofahrzeuge/elektrofahrzeuge/foerderung.php",
              "Comments": null,
              "PhonePrimaryContact": " 03831 2410",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "service@stadtwerke-stralsund.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 153,
              "Title": "Stadtwerke Strahlsund"
            },
            {
              "WebsiteURL": "http://www.stadtwerke-uetersen.de/?page_id=1722",
              "Comments": null,
              "PhonePrimaryContact": " +49 4122 92786 80",
              "PhoneSecondaryContact": "04122 / 9 27 86 82",
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "service@stadtwerke-uetersen.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 169,
              "Title": "Stadtwerke Uetersen"
            },
            {
              "WebsiteURL": "http://www.stadtwerke-verden.de/privatkunden/energiedienstleistungen/elektromobilitaet.html",
              "Comments": null,
              "PhonePrimaryContact": "04231 915112",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 134,
              "Title": "Stadtwerke Verden"
            },
            {
              "WebsiteURL": "http://www.stadtwerke-wernigerode.de/published_page.aspx?id=132",
              "Comments": null,
              "PhonePrimaryContact": "03943  556-111",
              "PhoneSecondaryContact": "(03943) 556-270 ",
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "stefan.streichert@stadtwerke-wernigerode.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 130,
              "Title": "Stadtwerke Wernigerode"
            },
            {
              "WebsiteURL": "http://stadtwerke.wittenberg.de/",
              "Comments": null,
              "PhonePrimaryContact": " 03491 4700",
              "PhoneSecondaryContact": "0800 759 0800",
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "slw@stadtwerke.wittenberg.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 159,
              "Title": "Stadtwerke Wittenberg"
            },
            {
              "WebsiteURL": "http://www.stadtwerke-wf.de/",
              "Comments": null,
              "PhonePrimaryContact": "(0 53 31) 4 08 - 1 14 ",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "service@stadtwerke-wf.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 166,
              "Title": "Stadtwerke Wolfenbttel"
            },
            {
              "WebsiteURL": "http://www.sgcc.com.cn/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3290,
              "Title": "State Grid Corporation of China (SGCC)"
            },
            {
              "WebsiteURL": "http://www.stromnetz.hamburg/ueber-uns/innovationen/e-mobility/",
              "Comments": null,
              "PhonePrimaryContact": "+49 40 4920200",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@stromnetz-hamburg.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 220,
              "Title": "Stromnetz Hamburg"
            },
            {
              "WebsiteURL": "http://stromticket.de/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 160,
              "Title": "StromTicket"
            },
            {
              "WebsiteURL": "http://www.stromtreter.de/",
              "Comments": null,
              "PhonePrimaryContact": " +49 800 7333532",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3269,
              "Title": "Stromtreter"
            },
            {
              "WebsiteURL": "http://www.stromtreter.de/",
              "Comments": null,
              "PhonePrimaryContact": " +49 800 7333532",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3270,
              "Title": "Stromtreter"
            },
            {
              "WebsiteURL": "http://suncountryhighway.ca/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 51,
              "Title": "Sun Country Highway"
            },
            {
              "WebsiteURL": "http://www.sun-stadtwerke.de/energien-der-zukunft/e-mobilitaet/ladekarte.html",
              "Comments": null,
              "PhonePrimaryContact": "0800 0 22 88 44 ",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "sun@sun-stadtwerke.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 138,
              "Title": "SUN Stadtwerke Union Nordhessen"
            },
            {
              "WebsiteURL": "https://www.swb-gruppe.de/verantwortung/swb-und-umwelt/fahren-mit-strom.php",
              "Comments": null,
              "PhonePrimaryContact": " 0421 3590",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 125,
              "Title": "SWB / EWE"
            },
            {
              "WebsiteURL": "https://swisscharge.ch/",
              "Comments": null,
              "PhonePrimaryContact": "071 388 11 50",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@swisscharge.ch",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3273,
              "Title": "Swisscharge (CH)"
            },
            {
              "WebsiteURL": "http://www.syane.orios-infos.com/",
              "Comments": "Syndicat des nergies et de l'amngement numrique de la Haute-Savoie",
              "PhonePrimaryContact": "+33 4 50 33 50 60",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3277,
              "Title": "Syane"
            },
            {
              "WebsiteURL": "http://tank.rast.de/emobility/",
              "Comments": null,
              "PhonePrimaryContact": "+49 228 9220",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "kundenservice@tank.rast.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3248,
              "Title": "Tank & Rast"
            },
            {
              "WebsiteURL": "http://www.tanke-wienenergie.at/",
              "Comments": null,
              "PhonePrimaryContact": "0800 510 820",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 142,
              "Title": "TANKE Wien Energie"
            },
            {
              "WebsiteURL": "http://www.teslamotors.com",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 23,
              "Title": "Tesla Motors (Worldwide)"
            },
            {
              "WebsiteURL": "https://www.chargepointgenie.com/",
              "Comments": null,
              "PhonePrimaryContact": "+44(0)20 3598 4087",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "GenieSupport@chargepointservices.com",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 150,
              "Title": "The GeniePoint Network ( ChargePoint Services )"
            },
            {
              "WebsiteURL": "http://www.thenewmotion.be/",
              "Comments": null,
              "PhonePrimaryContact": " +32 (0)2 588 1251",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@thenewmotion.be",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 157,
              "Title": "The New Motion (BE)"
            },
            {
              "WebsiteURL": "http://www.thenewmotion.de/",
              "Comments": null,
              "PhonePrimaryContact": "030 215 028 48",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@thenewmotion.de",
              "FaultReportEmail": "service@thenewmotion.de",
              "IsRestrictedEdit": false,
              "ID": 156,
              "Title": "The New Motion (DE)"
            },
            {
              "WebsiteURL": "http://www.thenewmotion.com/",
              "Comments": null,
              "PhonePrimaryContact": "088 01 09 500",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@thenewmotion.com",
              "FaultReportEmail": "klantenservice@thenewmotion.com",
              "IsRestrictedEdit": false,
              "ID": 47,
              "Title": "The New Motion (NL)"
            },
            {
              "WebsiteURL": "http://therevproject.com/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "rev@therevproject.com",
              "FaultReportEmail": "rev@therevproject.com",
              "IsRestrictedEdit": false,
              "ID": 57,
              "Title": "The REV Project (UWA - Australia)"
            },
            {
              "WebsiteURL": "http://www.theplugincompany.com/en/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 35,
              "Title": "ThePluginCompany (Belgium)"
            },
            {
              "WebsiteURL": "https://www.tiwag.at/no_cache/privatkunden/energieeffizienz/mobilitaet/",
              "Comments": null,
              "PhonePrimaryContact": "0800 818 819",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 18,
              "Title": "TIWAG Tiroler Wasserkraft AG (AT)"
            },
            {
              "WebsiteURL": "http://www.total.be/fr/carburants/carburant-station/plug-to-drive/finder-plugtodrive.html",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 175,
              "Title": "TOTAL Be PlugToDrive"
            },
            {
              "WebsiteURL": "http://www.total.nl/services-tankstations/plug-to-drive.html",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 174,
              "Title": "TOTAL Nl PlugToDrive "
            },
            {
              "WebsiteURL": "https://www.tsk.sk/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3306,
              "Title": "Tren?iansky samosprvny kraj"
            },
            {
              "WebsiteURL": "https://www.ubitricity.com/",
              "Comments": null,
              "PhonePrimaryContact": "+49 (0)30 398 371 690",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "contact@ubitricity.com",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 2244,
              "Title": "ubitricity"
            },
            {
              "WebsiteURL": "http://www.ultra-fast.nl/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 64,
              "Title": "Ultra-Fast"
            },
            {
              "WebsiteURL": "http://www.umeaenergi.se/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "umea.energi@umeaenergi.se",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3249,
              "Title": "Ume Energi "
            },
            {
              "WebsiteURL": "https://www.urbener.com/",
              "Comments": null,
              "PhonePrimaryContact": "+34 976 29 89 84",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "urbener@urbener.com",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3278,
              "Title": "Urbener Energa"
            },
            {
              "WebsiteURL": "http://www.useda.fr/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3256,
              "Title": "USEDA (FR)"
            },
            {
              "WebsiteURL": "http://www.uewag.de/energie/stromtankstellen",
              "Comments": null,
              "PhonePrimaryContact": "0661 12-100",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 65,
              "Title": "UWAG"
            },
            {
              "WebsiteURL": "http://www.vattenfall.de/de/emobility/emobility.htm",
              "Comments": null,
              "PhonePrimaryContact": "0800 - 2 335 335",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "emobility@vattenfall.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 108,
              "Title": "Vattenfall"
            },
            {
              "WebsiteURL": "https://vector.co.nz/evcharging",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 233,
              "Title": "Vector (NZ)"
            },
            {
              "WebsiteURL": "http://www.versorgungsbetriebe.de/hannmuendenGips/Gips?SessionMandant=HannMuenden&Anwendung=CMSWebpage&Methode=ShowHTMLAusgabe&RessourceID=12045",
              "Comments": null,
              "PhonePrimaryContact": " 01802 707800",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@versorgungsbetriebe.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 144,
              "Title": "Versorgungsbetriebe Hann. Mnden"
            },
            {
              "WebsiteURL": "http://www.emobility-vibrate.eu/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@emobility-vibrate.eu",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 231,
              "Title": "Vibrate"
            },
            {
              "WebsiteURL": "https://www.viesgo.com/es/nuevas-tecnologias-e-innovacion/movilidad-electrica/puntos-de-recarga-publica",
              "Comments": "Formerly known as E.ON Energa",
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 1237,
              "Title": "Viesgo (Spain)"
            },
            {
              "WebsiteURL": "http://virta.fi/",
              "Comments": null,
              "PhonePrimaryContact": "0800-02200",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 136,
              "Title": "VIRTA"
            },
            {
              "WebsiteURL": "http://vitaemobility.com/",
              "Comments": null,
              "PhonePrimaryContact": "+32 473 94 98 27",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@vitaemobility.eu",
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 61,
              "Title": "Vitaemobility"
            },
            {
              "WebsiteURL": "http://www.vlotte.at/",
              "Comments": null,
              "PhonePrimaryContact": "+43 5574 9000 ",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "elektromobil@vkw.at",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 99,
              "Title": "Vlotte"
            },
            {
              "WebsiteURL": "http://www.voltacharging.com",
              "Comments": null,
              "PhonePrimaryContact": "  1-888-264-2208",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@voltacharging.com",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 50,
              "Title": "Volta Charging"
            },
            {
              "WebsiteURL": "http://www.schneller-strom-tanken.de/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 126,
              "Title": "VR Schneller-Strom-tanken"
            },
            {
              "WebsiteURL": "http://www.zelenabuducnost.sk/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 193,
              "Title": "VSE"
            },
            {
              "WebsiteURL": "https://www.wel.co.nz/projects/electric-vehicles/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 28,
              "Title": "WEL Networks (NZ)"
            },
            {
              "WebsiteURL": "http://www.werraenergie.de/www/werraenergie/webinfo/webinfo.nsf/DocsID/E-Mobilitaet",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 195,
              "Title": "Werraenergie"
            },
            {
              "WebsiteURL": "http://westarelectrogo.com/",
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 34,
              "Title": "Westar Energy ElectroGo"
            },
            {
              "WebsiteURL": "http://www.wirtschaftsbetriebe-nienburg.de/",
              "Comments": null,
              "PhonePrimaryContact": "(05021) 87-315 ",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "wirtschaftsbetriebe@nienburg.de",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 183,
              "Title": "Wirtschaftsbetriebe Stadt Nienburg"
            },
            {
              "WebsiteURL": null,
              "Comments": "Energy company based in Argentina",
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3300,
              "Title": "YPF"
            },
            {
              "WebsiteURL": "http://www.zeag-energie.de",
              "Comments": null,
              "PhonePrimaryContact": "07131 610-0",
              "PhoneSecondaryContact": "07131 56-4248",
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 176,
              "Title": "ZEAG Energie"
            },
            {
              "WebsiteURL": "http://www.ze-mo.be/",
              "Comments": null,
              "PhonePrimaryContact": "+32 10 750 650",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@ze-mo.be",
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 54,
              "Title": "ZE-MO (Be)"
            },
            {
              "WebsiteURL": "https://www.zencar.eu/",
              "Comments": null,
              "PhonePrimaryContact": "+32 2 669 77 91",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@zencar.eu",
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 100,
              "Title": "ZenCar"
            },
            {
              "WebsiteURL": "http://zerocarbonworld.org/",
              "Comments": "UK Charity promoting carbon reduction",
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": "info@zerocarbonworld.com",
              "FaultReportEmail": "info@zerocarbonworld.com",
              "IsRestrictedEdit": null,
              "ID": 19,
              "Title": "Zero Carbon World"
            },
            {
              "WebsiteURL": "https://www.zse.sk/sk/Uvod/Firmy-a-organizacie/Nabijacie-stanice",
              "Comments": "Energy company in Slovakia",
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3259,
              "Title": "ZSE (SK)"
            },
            {
              "WebsiteURL": "https://www.moesk.ru/spec_projects/moesk_ev/",
              "Comments": null,
              "PhonePrimaryContact": "8 (800) 700-40-70",
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": false,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": false,
              "ID": 3301,
              "Title": "????? (Russia)"
            }
          ],
          "StatusTypes": [
            {
              "IsOperational": null,
              "IsUserSelectable": true,
              "ID": 0,
              "Title": "Unknown"
            },
            {
              "IsOperational": true,
              "IsUserSelectable": false,
              "ID": 10,
              "Title": "Currently Available (Automated Status)"
            },
            {
              "IsOperational": true,
              "IsUserSelectable": false,
              "ID": 20,
              "Title": "Currently In Use (Automated Status)"
            },
            {
              "IsOperational": true,
              "IsUserSelectable": true,
              "ID": 30,
              "Title": "Temporarily Unavailable"
            },
            {
              "IsOperational": true,
              "IsUserSelectable": true,
              "ID": 50,
              "Title": "Operational"
            },
            {
              "IsOperational": true,
              "IsUserSelectable": true,
              "ID": 75,
              "Title": "Partly Operational (Mixed)"
            },
            {
              "IsOperational": false,
              "IsUserSelectable": true,
              "ID": 100,
              "Title": "Not Operational"
            },
            {
              "IsOperational": false,
              "IsUserSelectable": true,
              "ID": 150,
              "Title": "Planned For Future Date"
            },
            {
              "IsOperational": false,
              "IsUserSelectable": true,
              "ID": 200,
              "Title": "Removed (Decomissioned)"
            }
          ],
          "SubmissionStatusTypes": [
            {
              "IsLive": false,
              "ID": 1,
              "Title": "Submission Under Review"
            },
            {
              "IsLive": false,
              "ID": 50,
              "Title": "Imported and Under Review"
            },
            {
              "IsLive": true,
              "ID": 100,
              "Title": "Imported and Published"
            },
            {
              "IsLive": true,
              "ID": 200,
              "Title": "Submission Published"
            },
            {
              "IsLive": false,
              "ID": 250,
              "Title": "Submission Rejected - Incomplete"
            },
            {
              "IsLive": false,
              "ID": 1000,
              "Title": "Delisted"
            },
            {
              "IsLive": false,
              "ID": 1001,
              "Title": "Delisted - Duplicate"
            },
            {
              "IsLive": false,
              "ID": 1002,
              "Title": "Delisted - No Longer Active"
            },
            {
              "IsLive": false,
              "ID": 1005,
              "Title": "Delisted - Superseded by Update"
            },
            {
              "IsLive": false,
              "ID": 1010,
              "Title": "Delisted - Not Public Information/Is Private Residence"
            },
            {
              "IsLive": false,
              "ID": 1020,
              "Title": "Delisted - Removed by Import Data Provider"
            }
          ],
          "UsageTypes": [
            {
              "IsPayAtLocation": null,
              "IsMembershipRequired": null,
              "IsAccessKeyRequired": null,
              "ID": 0,
              "Title": "(Unknown)"
            },
            {
              "IsPayAtLocation": false,
              "IsMembershipRequired": false,
              "IsAccessKeyRequired": false,
              "ID": 6,
              "Title": "Private - For Staff, Visitors or Customers"
            },
            {
              "IsPayAtLocation": null,
              "IsMembershipRequired": true,
              "IsAccessKeyRequired": null,
              "ID": 2,
              "Title": "Private - Restricted Access"
            },
            {
              "IsPayAtLocation": null,
              "IsMembershipRequired": null,
              "IsAccessKeyRequired": null,
              "ID": 3,
              "Title": "Privately Owned - Notice Required"
            },
            {
              "IsPayAtLocation": null,
              "IsMembershipRequired": null,
              "IsAccessKeyRequired": null,
              "ID": 1,
              "Title": "Public"
            },
            {
              "IsPayAtLocation": false,
              "IsMembershipRequired": true,
              "IsAccessKeyRequired": true,
              "ID": 4,
              "Title": "Public - Membership Required"
            },
            {
              "IsPayAtLocation": false,
              "IsMembershipRequired": false,
              "IsAccessKeyRequired": false,
              "ID": 7,
              "Title": "Public - Notice Required"
            },
            {
              "IsPayAtLocation": true,
              "IsMembershipRequired": false,
              "IsAccessKeyRequired": false,
              "ID": 5,
              "Title": "Public - Pay At Location"
            }
          ],
          "UserCommentTypes": [
            {
              "ID": 10,
              "Title": "General Comment"
            },
            {
              "ID": 50,
              "Title": "Important Notice (For Other Users)"
            },
            {
              "ID": 100,
              "Title": "Suggested Change (Note To Editors)"
            },
            {
              "ID": 110,
              "Title": "Suggested Changed (Actioned By Editor)"
            },
            {
              "ID": 1000,
              "Title": "Fault Report (Notice To Users And Operator)"
            }
          ],
          "CheckinStatusTypes": [
            {
              "IsPositive": null,
              "IsAutomatedCheckin": false,
              "ID": 0,
              "Title": "Did Not Visit Location"
            },
            {
              "IsPositive": true,
              "IsAutomatedCheckin": false,
              "ID": 10,
              "Title": "Charged Successfully"
            },
            {
              "IsPositive": true,
              "IsAutomatedCheckin": true,
              "ID": 15,
              "Title": "Charged Successfully (Automated Checkin)"
            },
            {
              "IsPositive": false,
              "IsAutomatedCheckin": false,
              "ID": 20,
              "Title": "Failed to Charge (Equipment Not Operational)"
            },
            {
              "IsPositive": false,
              "IsAutomatedCheckin": false,
              "ID": 22,
              "Title": "Failed to Charge (Equipment Not Fully Installed)"
            },
            {
              "IsPositive": false,
              "IsAutomatedCheckin": false,
              "ID": 25,
              "Title": "Failed to Charge (Equipment Problem)"
            },
            {
              "IsPositive": false,
              "IsAutomatedCheckin": false,
              "ID": 30,
              "Title": "Failed to Charge (Equipment Not Compatible)"
            },
            {
              "IsPositive": false,
              "IsAutomatedCheckin": false,
              "ID": 40,
              "Title": "Failed to Charge (Required Other Access Card/Fob etc.)"
            },
            {
              "IsPositive": false,
              "IsAutomatedCheckin": false,
              "ID": 50,
              "Title": "Failed to Charge (No Charging Equipment Present)"
            },
            {
              "IsPositive": true,
              "IsAutomatedCheckin": false,
              "ID": 100,
              "Title": "Charging Spot In Use (Other EV Parked)"
            },
            {
              "IsPositive": false,
              "IsAutomatedCheckin": false,
              "ID": 110,
              "Title": "Charging Spot In Use (Non-EV Parked)"
            },
            {
              "IsPositive": false,
              "IsAutomatedCheckin": false,
              "ID": 120,
              "Title": "Charging Spot Not Accessible (Access locked or site closed)"
            },
            {
              "IsPositive": false,
              "IsAutomatedCheckin": false,
              "ID": 130,
              "Title": "Charging Spot Not Found (Inadequate or Incorrect Details)"
            },
            {
              "IsPositive": true,
              "IsAutomatedCheckin": false,
              "ID": 140,
              "Title": "Equipment & Location Confirmed Correct"
            },
            {
              "IsPositive": false,
              "IsAutomatedCheckin": false,
              "ID": 150,
              "Title": "Location Is A Duplicate"
            },
            {
              "IsPositive": false,
              "IsAutomatedCheckin": false,
              "ID": 160,
              "Title": "Equipment/Location Has Been Decommissioned"
            },
            {
              "IsPositive": false,
              "IsAutomatedCheckin": false,
              "ID": 200,
              "Title": "Other (Negative/Bad)"
            },
            {
              "IsPositive": true,
              "IsAutomatedCheckin": false,
              "ID": 210,
              "Title": "Other (Positive/Good)"
            }
          ],
          "DataTypes": [
            {
              "ID": 1,
              "Title": "Single Line Text"
            },
            {
              "ID": 2,
              "Title": "Multi Line Text"
            },
            {
              "ID": 3,
              "Title": "Number"
            },
            {
              "ID": 4,
              "Title": "Date"
            },
            {
              "ID": 5,
              "Title": "Boolean"
            },
            {
              "ID": 10,
              "Title": "Option List"
            }
          ],
          "MetadataGroups": [
            {
              "DataProviderID": 1,
              "IsRestrictedEdit": false,
              "IsPublicInterest": true,
              "MetadataFields": [
                {
                  "MetadataGroupID": 1,
                  "DataTypeID": 10,
                  "DataType": null,
                  "MetadataFieldOptions": [
                    {
                      "MetadataFieldID": 1,
                      "ID": 1,
                      "Title": "Shopping"
                    },
                    {
                      "MetadataFieldID": 1,
                      "ID": 2,
                      "Title": "Hotel"
                    },
                    {
                      "MetadataFieldID": 1,
                      "ID": 3,
                      "Title": "Restaurant"
                    },
                    {
                      "MetadataFieldID": 1,
                      "ID": 8,
                      "Title": "Parking"
                    },
                    {
                      "MetadataFieldID": 1,
                      "ID": 9,
                      "Title": "Secure"
                    }
                  ],
                  "ID": 1,
                  "Title": "POI Type"
                },
                {
                  "MetadataGroupID": 1,
                  "DataTypeID": 1,
                  "DataType": null,
                  "MetadataFieldOptions": null,
                  "ID": 2,
                  "Title": "Access Hours"
                },
                {
                  "MetadataGroupID": 1,
                  "DataTypeID": 10,
                  "DataType": null,
                  "MetadataFieldOptions": [
                    {
                      "MetadataFieldID": 4,
                      "ID": 4,
                      "Title": "Positioning"
                    }
                  ],
                  "ID": 4,
                  "Title": "Attribution"
                },
                {
                  "MetadataGroupID": 1,
                  "DataTypeID": 10,
                  "DataType": null,
                  "MetadataFieldOptions": [
                    {
                      "MetadataFieldID": 5,
                      "ID": 5,
                      "Title": "Car"
                    },
                    {
                      "MetadataFieldID": 5,
                      "ID": 6,
                      "Title": "Motorcycle"
                    },
                    {
                      "MetadataFieldID": 5,
                      "ID": 7,
                      "Title": "Bus"
                    },
                    {
                      "MetadataFieldID": 5,
                      "ID": 10,
                      "Title": "Goods/Delivery Vehicles"
                    },
                    {
                      "MetadataFieldID": 5,
                      "ID": 11,
                      "Title": "Taxi"
                    },
                    {
                      "MetadataFieldID": 5,
                      "ID": 12,
                      "Title": "Bike"
                    }
                  ],
                  "ID": 5,
                  "Title": "Vehicle Type"
                }
              ],
              "ID": 1,
              "Title": "Extended Information"
            },
            {
              "DataProviderID": 1,
              "IsRestrictedEdit": false,
              "IsPublicInterest": false,
              "MetadataFields": [
                {
                  "MetadataGroupID": 2,
                  "DataTypeID": 10,
                  "DataType": null,
                  "MetadataFieldOptions": null,
                  "ID": 3,
                  "Title": "Group of Companies"
                }
              ],
              "ID": 2,
              "Title": "Supporting Data"
            }
          ],
          "UserProfile": null,
          "ChargePoint": {
            "ID": -1,
            "UUID": "7755082a-b92e-486c-8df8-32b85458eaef",
            "ParentChargePointID": null,
            "DataProviderID": null,
            "DataProvider": {
              "WebsiteURL": null,
              "Comments": null,
              "DataProviderStatusType": null,
              "IsRestrictedEdit": false,
              "IsOpenDataLicensed": null,
              "IsApprovedImport": null,
              "License": null,
              "DateLastImported": null,
              "ID": 0,
              "Title": null
            },
            "DataProvidersReference": null,
            "OperatorID": null,
            "OperatorInfo": {
              "WebsiteURL": null,
              "Comments": null,
              "PhonePrimaryContact": null,
              "PhoneSecondaryContact": null,
              "IsPrivateIndividual": null,
              "AddressInfo": null,
              "BookingURL": null,
              "ContactEmail": null,
              "FaultReportEmail": null,
              "IsRestrictedEdit": null,
              "ID": 0,
              "Title": null
            },
            "OperatorsReference": null,
            "UsageTypeID": null,
            "UsageType": {
              "IsPayAtLocation": null,
              "IsMembershipRequired": null,
              "IsAccessKeyRequired": null,
              "ID": 0,
              "Title": null
            },
            "UsageCost": null,
            "AddressInfo": {
              "ID": 0,
              "Title": null,
              "AddressLine1": null,
              "AddressLine2": null,
              "Town": null,
              "StateOrProvince": null,
              "Postcode": null,
              "CountryID": null,
              "Country": null,
              "Latitude": 0.0,
              "Longitude": 0.0,
              "ContactTelephone1": null,
              "ContactTelephone2": null,
              "ContactEmail": null,
              "AccessComments": null,
              "RelatedURL": null,
              "Distance": null,
              "DistanceUnit": 0
            },
            "NumberOfPoints": 1,
            "GeneralComments": "",
            "DatePlanned": null,
            "DateLastConfirmed": "2018-05-20T17:42:06.337Z",
            "StatusTypeID": null,
            "StatusType": {
              "IsOperational": null,
              "IsUserSelectable": false,
              "ID": 0,
              "Title": null
            },
            "DateLastStatusUpdate": "2018-05-20T17:42:06.337Z",
            "DataQualityLevel": 1,
            "DateCreated": "2018-05-20T17:42:06.337Z",
            "SubmissionStatusTypeID": null,
            "SubmissionStatus": null,
            "UserComments": null,
            "PercentageSimilarity": null,
            "Connections": [
              {
                "ID": 0,
                "ConnectionTypeID": null,
                "ConnectionType": null,
                "Reference": null,
                "StatusTypeID": null,
                "StatusType": null,
                "LevelID": null,
                "Level": null,
                "Amps": null,
                "Voltage": null,
                "PowerKW": null,
                "CurrentTypeID": null,
                "CurrentType": null,
                "Quantity": null,
                "Comments": null
              }
            ],
            "MediaItems": null,
            "MetadataValues": null,
            "IsRecentlyVerified": true,
            "DateLastVerified": "2018-05-20T17:42:06.337Z"
          },
          "UserComment": {
            "ID": 0,
            "ChargePointID": 0,
            "CommentTypeID": null,
            "CommentType": {
              "ID": 10,
              "Title": "General Comment"
            },
            "UserName": null,
            "Comment": "",
            "Rating": null,
            "RelatedURL": null,
            "DateCreated": "2018-05-20T17:42:06.337Z",
            "User": null,
            "CheckinStatusTypeID": null,
            "CheckinStatusType": {
              "IsPositive": null,
              "IsAutomatedCheckin": false,
              "ID": 0,
              "Title": "Did Not Visit Location"
            }
          }
        }''')
        instance = Ocmpy()

        # Act
        parseresult = instance._parsereferencedata(testdata)

        # Assert
        self.assertTrue(parseresult)
        self.assertEqual(len(instance._chargertypesref), 3)
        self.assertEqual(len(instance._connectiontypesref), 39)
        self.assertEqual(len(instance._currenttypesref), 3)
        self.assertEqual(len(instance._countriesref), 249)
        self.assertEqual(len(instance._dataprovidersref), 18)
        self.assertEqual(len(instance._operatorsref), 308)
        self.assertEqual(len(instance._statustypesref), 9)
        self.assertEqual(len(instance._submissionstatustypesref), 11)
        self.assertEqual(len(instance._usagetypesref), 8)
        self.assertEqual(len(instance._usercommenttypesref), 5)
        self.assertEqual(len(instance._checkinstatustypesref), 18)
        self.assertEqual(len(instance._datatypesref), 6)
        self.assertEqual(len(instance._metadatagroupsref), 2)

    def test_ParseInvalidReferenceData(self):
        # Arrange
        testdata = loads(r'''{
                  "ChargerTypes": [
                    {
                      "Comments": "Under 2 kW, usually domestic socket types",
                      "IsFastChargeCapable": false,
                      "ID": 1,
                      "Title": "Level 1 : Low (Under 2kW)"
                    },
                    {
                      "Comments": "Over 2 kW, usually non-domestic socket type",
                      "IsFastChargeCapable": false,
                      "ID": 2,
                      "Title": "Level 2 : Medium (Over 2kW)"
                    },
                    {
                      "Comments": "40KW and Higher",
                      "IsFastChargeCapable": true,
                      "ID": 3,
                      "Title": "Level 3:  High (Over 40kW)"
                    }
                  ]
                }''')
        instance = Ocmpy()
        func = instance._parsereferencedata

        # Act

        # Assert
        self.assertRaises(OcmpyLoadReferenceDataFailed, func, testdata)

    def test_ParseLiveReferenceData(self):
        # Arrange
        testdata = get('https://api.openchargemap.io/v2/referencedata/').json()
        instance = Ocmpy()
        exception = None

        # Act
        try:
            instance._parsereferencedata(testdata)
        except:
            exception = exc_info()[0]

        # Assert
        self.assertIsNone(exception)


class BuildUrlTests(TestCase):
    def test__BuildValidUrl(self):
        # Arrange
        instance = Ocmpy()
        testpath = 'api/'
        testargs = {'arg1': 'val1'}
        baseurl = 'https://api.openchargemap.io/v2/'

        # Act
        testurl = instance._buildurl(testpath, testargs)

        # Assert
        self.assertEqual(testurl, 'https://api.openchargemap.io/v2/api/?arg1=val1')


class OcmpyLookupTests(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.instance = Ocmpy()
        cls.instance.open()

    def test_LookupValidOperator(self):
        # Arrange
        operatorid = 1

        # Act
        operator = self.instance.lookupoperatorbyid(operatorid)

        # Assert
        self.assertEqual(operator['Title'], '(Unknown Operator)')

    def test_LookupInvalidOperator(self):
        # Arrange
        operatorid = 999999
        func = self.instance.lookupoperatorbyid

        # Act & Assert
        self.assertRaises(OcmpyOperatorNotFound, func, operatorid)

    def test_LookupValidDataProvider(self):
        # Arrange
        dataproviderid = 1

        # Act
        dataprovider = self.instance.lookupdataproviderbyid(dataproviderid)

        # Assert
        self.assertEqual(dataprovider['Title'], 'Open Charge Map Contributors')

    def test_LookupInvalidDataProvider(self):
        # Arrange
        dataproviderid = 999999
        func = self.instance.lookupdataproviderbyid

        # Act & Assert
        self.assertRaises(OcmpyDataProviderNotFound, func, dataproviderid)

    def test_LookupValidUsageType(self):
        # Arrange
        usagetypeid = 1

        # Act
        usagetype = self.instance.lookupusagetypebyid(usagetypeid)

        # Assert
        self.assertEqual(usagetype['Title'], 'Public')

    def test_LookupInvalidUsageType(self):
        # Arrange
        usagetypeid = 999999
        func = self.instance.lookupusagetypebyid

        # Act & Assert
        self.assertRaises(OcmpyUsageTypeNotFound, func, usagetypeid)

    def test_LookupValidCountry(self):
        # Arrange
        countryid = 1

        # Act
        country = self.instance.lookupcountrybyid(countryid)

        # Assert
        self.assertEqual(country['Title'], 'United Kingdom')

    def test_LookupInvalidCountry(self):
        # Arrange
        countryid = 999999
        func = self.instance.lookupcountrybyid

        # Act & Assert
        self.assertRaises(OcmpyCountryNotFound, func, countryid)

    def test_LookupValidStatusType(self):
        # Arrange
        statustypeid = 10

        # Act
        statustype = self.instance.lookupstatustypebyid(statustypeid)

        # Assert
        self.assertEqual(statustype['Title'], 'Currently Available (Automated Status)')

    def test_LookupInvalidStatusType(self):
        # Arrange
        countryid = 999999
        func = self.instance.lookupstatustypebyid

        # Act & Assert
        self.assertRaises(OcmpyStatusTypeNotFound, func, countryid)

    def test_LookupValidSubmissionStatusType(self):
        # Arrange
        submissionstatustypeid = 1

        # Act
        submissionstatustype = self.instance.lookupsubmissionstatustypebyid(submissionstatustypeid)

        # Assert
        self.assertEqual(submissionstatustype['Title'], 'Submission Under Review')

    def test_LookupInvalidSubmissionStatusType(self):
        # Arrange
        submissionstatustypeid = 999999
        func = self.instance.lookupsubmissionstatustypebyid

        # Act & Assert
        self.assertRaises(OcmpySubmissionStatusTypeNotFound, func, submissionstatustypeid)

    def test_LookupValidLevel(self):
        # Arrange
        levelid = 1

        # Act
        country = self.instance.lookuplevelbyid(levelid)

        # Assert
        self.assertEqual(country['Title'], 'Level 1 : Low (Under 2kW)')

    def test_LookupInvalidLevel(self):
        # Arrange
        levelid = 999999
        func = self.instance.lookuplevelbyid

        # Act & Assert
        self.assertRaises(OcmpyLevelNotFound, func, levelid)

    def test_LookupValidConnectionType(self):
        # Arrange
        connectiontypeid = 1

        # Act
        country = self.instance.lookupconnectiontypebyid(connectiontypeid)

        # Assert
        self.assertEqual(country['Title'], 'J1772')

    def test_LookupInvalidConnectionType(self):
        # Arrange
        connectiontypeid = 999999
        func = self.instance.lookupconnectiontypebyid

        # Act & Assert
        self.assertRaises(OcmpyConnectionTypeNotFound, func, connectiontypeid)

    def test_LookupValidCurrentType(self):
        # Arrange
        currenttypeid = 10

        # Act
        country = self.instance.lookupcurrenttypebyid(currenttypeid)

        # Assert
        self.assertEqual(country['Title'], 'AC (Single-Phase)')

    def test_LookupInvalidCurrentType(self):
        # Arrange
        currenttypeid = 999999
        func = self.instance.lookupcurrenttypebyid

        # Act & Assert
        self.assertRaises(OcmpyCurrentTypeNotFound, func, currenttypeid)


class OcmpyStationTests(TestCase):
    def test_ParseValidStationData_internal(self):
        # Arrange
        testdata = loads('{"ID":104520,"UUID":"C9777FB8-A7F9-48A1-8FFC-4DF6BA5C0AAE","DataProviderID":1,'
                         '"OperatorID":23,"OperatorsReference":"37188","UsageTypeID":4,"AddressInfo":'
                         '{"ID":104866,"Title":"Pupatella","AddressLine1":"1 N Morris St","Town":"Richmond"'
                         ',"StateOrProvince":"VA","Postcode":"23220","CountryID":2,"Latitude":37.546347,'
                         '"Longitude":-77.456361,"ContactTelephone1":"(804) 355-3135","DistanceUnit":0},'
                         '"NumberOfPoints":2,"GeneralComments":"2 Tesla Connectors, up to 12kW.Available for'
                         ' customers. Self park.","StatusTypeID":50,"DateLastStatusUpdate":"2018-05-21T05:29:00Z",'
                         '"DataQualityLevel":1,"DateCreated":"2018-05-21T05:29:00Z","SubmissionStatusTypeID":200,'
                         '"Connections":[{"ID":147526,"ConnectionTypeID":30,"StatusTypeID":50,"LevelID":2,'
                         '"PowerKW":12.0,"CurrentTypeID":20,"Quantity":2}],"IsRecentlyVerified":true,'
                         '"DateLastVerified":"2018-05-21T05:29:00Z"}')

        # Act
        station = OcmpyStation(testdata)

        # Assert
        self.assertEqual(station._id, 104520)
        self.assertEqual(station._uuid, 'C9777FB8-A7F9-48A1-8FFC-4DF6BA5C0AAE')
        self.assertEqual(station._dataproviderid, 1)
        self.assertEqual(station._operatorid, 23)
        self.assertEqual(station._operatorsreference, '37188')
        self.assertEqual(station._usagetypeid, 4)
        self.assertEqual(station._addressid, 104866)
        self.assertEqual(station._addresstitle, 'Pupatella')
        self.assertEqual(station._addressline1, '1 N Morris St')
        self.assertEqual(station._addressline2, None)
        self.assertEqual(station._addresstown, 'Richmond')
        self.assertEqual(station._addressstateorprovince, 'VA')
        self.assertEqual(station._addresspostcode, '23220')
        self.assertEqual(station._addresscountryid, 2)
        self.assertEqual(station._addresslatitude, 37.546347)
        self.assertEqual(station._addresslongitude, -77.456361)
        self.assertEqual(station._addresscontacttelephone1, '(804) 355-3135')
        self.assertEqual(station._addressdistanceunit, 0)
        self.assertEqual(station._numberofpoints, 2)
        self.assertEqual(station._generalcomments, '2 Tesla Connectors, up to 12kW.Available for customers. Self park.')
        self.assertEqual(station._statustypeid, 50)
        self.assertEqual(station._datelaststatusupdate, '2018-05-21T05:29:00Z')
        self.assertEqual(station._dataqualitylevel, 1)
        self.assertEqual(station._datecreated, '2018-05-21T05:29:00Z')
        self.assertEqual(station._submissionstatustypeid, 200)
        self.assertEqual(station.Connections[0]._id, 147526)
        self.assertEqual(station.Connections[0]._connectiontypeid, 30)
        self.assertEqual(station.Connections[0]._statustypeid, 50)
        self.assertEqual(station.Connections[0]._levelid, 2)
        self.assertEqual(station.Connections[0]._powerkw, 12.0)
        self.assertEqual(station.Connections[0]._currenttypeid, 20)
        self.assertEqual(station.Connections[0]._quantity, 2)
        self.assertTrue(station._isrecentlyverified)
        self.assertEqual(station._datelastverified, '2018-05-21T05:29:00Z')

    def test_ParseValidStationData_public(self):
        # Arrange
        testdata = loads('{"ID":104520,"UUID":"C9777FB8-A7F9-48A1-8FFC-4DF6BA5C0AAE","DataProviderID":1,'
                         '"OperatorID":23,"OperatorsReference":"37188","UsageTypeID":4,"AddressInfo":'
                         '{"ID":104866,"Title":"Pupatella","AddressLine1":"1 N Morris St","Town":"Richmond"'
                         ',"StateOrProvince":"VA","Postcode":"23220","CountryID":2,"Latitude":37.546347,'
                         '"Longitude":-77.456361,"ContactTelephone1":"(804) 355-3135","DistanceUnit":0},'
                         '"NumberOfPoints":2,"GeneralComments":"2 Tesla Connectors, up to 12kW.Available for'
                         ' customers. Self park.","StatusTypeID":50,"DateLastStatusUpdate":"2018-05-21T05:29:00Z",'
                         '"DataQualityLevel":1,"DateCreated":"2018-05-21T05:29:00Z","SubmissionStatusTypeID":200,'
                         '"Connections":[{"ID":147526,"ConnectionTypeID":30,"StatusTypeID":50,"LevelID":2,'
                         '"PowerKW":12.0,"CurrentTypeID":20,"Quantity":2}],"IsRecentlyVerified":true,'
                         '"DateLastVerified":"2018-05-21T05:29:00Z"}')
        instance = Ocmpy()
        instance.open()

        # Act
        station = OcmpyStation(testdata, instance)

        # Assert
        self.assertEqual(station.ID, 104520)
        self.assertEqual(station.UUID, uuid.UUID('C9777FB8-A7F9-48A1-8FFC-4DF6BA5C0AAE'))
        self.assertEqual(station.DataProviderID, 1)
        self.assertEqual(station.DataProvider['Title'], 'Open Charge Map Contributors')
        self.assertEqual(station.OperatorID, 23)
        self.assertEqual(station.Operator['Title'], 'Tesla Motors (Worldwide)')
        self.assertEqual(station.OperatorsReference, '37188')
        self.assertEqual(station.UsageTypeID, 4)
        self.assertEqual(station.UsageType['Title'], 'Public - Membership Required')
        self.assertEqual(station.AddressID, 104866)
        self.assertEqual(station.Title, 'Pupatella')
        self.assertEqual(station.AddressLine1, '1 N Morris St')
        self.assertEqual(station.AddressLine2, None)
        self.assertEqual(station.Town, 'Richmond')
        self.assertEqual(station.State, 'VA')
        self.assertEqual(station.Province, 'VA')
        self.assertEqual(station.Postcode, '23220')
        self.assertEqual(station.CountryID, 2)
        self.assertEqual(station.Country['Title'], 'United States')
        self.assertEqual(station.PrintableAddress, ('Pupatella\n'
                                                    '1 N Morris St\n'
                                                    'Richmond, VA 23220\n'
                                                    'United States'))
        self.assertEqual(station.Latitude, 37.546347)
        self.assertEqual(station.Longitude, -77.456361)
        self.assertDictEqual(station.Coordinates, {'lat': 37.546347, 'lon': -77.456361})
        self.assertEqual(station.ContactPhone, '(804) 355-3135')
        self.assertEqual(station._addressdistanceunit, 0)
        self.assertEqual(station.NumberOfChargePoints, 2)
        self.assertEqual(station.GeneralComments, '2 Tesla Connectors, up to 12kW.Available for customers. Self park.')
        self.assertEqual(station.StatusTypeID, 50)
        self.assertEqual(station.Status['Title'], 'Operational')
        self.assertEqual(station.LastStatusUpdateDate, datetime.strptime('2018-05-21T05:29:00Z', '%Y-%m-%dT%H:%M:%SZ'))
        self.assertEqual(station.DataQualityLevel, 1)
        self.assertEqual(station.DateCreated, datetime.strptime('2018-05-21T05:29:00Z', '%Y-%m-%dT%H:%M:%SZ'))
        self.assertEqual(station.SubmissionStatusTypeID, 200)
        self.assertEqual(station.SubmissionStatus['Title'], 'Submission Published')
        self.assertEqual(station.Connections[0].ID, 147526)
        self.assertEqual(station.Connections[0].ConnectionTypeID, 30)
        self.assertEqual(station.Connections[0].ConnectionType['Title'], 'Tesla (Model S onwards)')
        self.assertEqual(station.Connections[0].StatusTypeID, 50)
        self.assertEqual(station.Connections[0].Status['Title'], 'Operational')
        self.assertEqual(station.Connections[0].LevelID, 2)
        self.assertEqual(station.Connections[0].Level['Title'], 'Level 2 : Medium (Over 2kW)')
        self.assertEqual(station.Connections[0].PowerKW, 12.0)
        self.assertEqual(station.Connections[0].CurrentTypeID, 20)
        self.assertEqual(station.Connections[0].Quantity, 2)
        self.assertTrue(station.RecentlyVerified)
        self.assertEqual(station.DateLastVerified, datetime.strptime('2018-05-21T05:29:00Z', '%Y-%m-%dT%H:%M:%SZ'))

    def test_ParseInvalidStationData(self):
        # Arrange
        # Data is missing ID field
        testdata = loads('{"UUID":"C9777FB8-A7F9-48A1-8FFC-4DF6BA5C0AAE","DataProviderID":1,'
                         '"OperatorID":23,"OperatorsReference":"37188","UsageTypeID":4,"AddressInfo":'
                         '{"ID":104866,"Title":"Pupatella","AddressLine1":"1 N Morris St","Town":"Richmond"'
                         ',"StateOrProvince":"VA","Postcode":"23220","CountryID":2,"Latitude":37.546347,'
                         '"Longitude":-77.456361,"ContactTelephone1":"(804) 355-3135","DistanceUnit":0},'
                         '"NumberOfPoints":2,"GeneralComments":"2 Tesla Connectors, up to 12kW.Available for'
                         ' customers. Self park.","StatusTypeID":50,"DateLastStatusUpdate":"2018-05-21T05:29:00Z",'
                         '"DataQualityLevel":1,"DateCreated":"2018-05-21T05:29:00Z","SubmissionStatusTypeID":200,'
                         '"Connections":[{"ID":147526,"ConnectionTypeID":30,"StatusTypeID":50,"LevelID":2,'
                         '"PowerKW":12.0,"CurrentTypeID":20,"Quantity":2}],"IsRecentlyVerified":true,'
                         '"DateLastVerified":"2018-05-21T05:29:00Z"}')

        # Assert
        self.assertRaises(OcmpyLoadStationFailed, OcmpyStation, testdata)

    def test_ParseLiveStationData(self):
        # Arrange
        testdata = get('https://api.openchargemap.io/v2/poi/'
                       '?output=json&countrycode=US&maxresults=100&compact=true&verbose=false').json()

        for test in testdata:
            # Act
            station = OcmpyStation(test)

            # Assert
            self.assertIsNotNone(station)


class OcmpyGetStationTests(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.instance = Ocmpy()
        cls.instance.open()

    def test_GetStations(self):
        # Arrange
        args = OcmpyArgs(latitude=40, longitude=-105, distance=100, maxresults=10)

        # Act
        stations = self.instance.getstations(args)

        # Assert
        self.assertIsNotNone(stations)
        self.assertIsNotNone(stations[0].ID)
        self.assertIsNotNone(stations[0].Country)
        self.assertIsNotNone(stations[0].Operator)
        self.assertIsNotNone(stations[0].DataProvider)
        self.assertIsNotNone(stations[0].UsageType)
        self.assertIsNotNone(stations[0].Status)
        self.assertIsNotNone(stations[0].SubmissionStatus)
        self.assertIsNotNone(stations[0].Connections)
        self.assertIsNotNone(stations[0].Connections[0].Level)
        self.assertIsNotNone(stations[0].Connections[0].ConnectionType)
        self.assertIsNotNone(stations[0].Connections[0].CurrentType)
